import os
import sqlite3
from base64 import b64encode, b64decode

import numpy as np

from sqlalchemy import MetaData, create_engine, func, text, and_
from sqlalchemy.sql import select
from sqlalchemy.orm import sessionmaker
from sqlalchemy.pool import SingletonThreadPool



__version__ = '1'

def make_engine(dbname):
    "create engine for sqlite connection"
    return create_engine('sqlite:///%s' % (dbname),
                         poolclass=SingletonThreadPool,
                         connect_args={'check_same_thread': False})

def isAMSCIFDB(dbname):
    """whether a file is a valid AMSCIF database

    Args:
        dbname (string): name of AMSCIF database file

    Returns:
        bool: is file a valid AMSCIF database

    Notes:
      1. must be a sqlite db file, with tables
        'cif', 'elements', 'spacegroup'
    """
    _tables = ('cif', 'elements', 'spacegroups')
    result = False
    try:
        engine = make_engine(dbname)
        meta = MetaData(engine)
        meta.reflect()
        result = all([t in meta.tables for t in _tables])
    except:
        pass
    return result


farray_scale = 4.e6

def encode_farray(dat):
    """encodes a list of fractional coordinate as strings (stricly on (-1,1))
    to an string for saving to db, to be decoded by  decode_farray()
    preserves precision to slightly better than 6 digits
    """
    work = []
    for d in dat:
        if d == '?':
            work.append(2) # out-of-bounds as '?'
        elif d == '.':
            work.append(3) # out-of-bounds as '.'
        else:
            if '(' in d or '(' in d:
                d = d.replace(')', ' : ').replace('(', ' : ')
                d = d.split(':')[0].strip()
            try:
                fval = float(d)
            except ValueError:
                d  = '0'
            work.append(d)
    x = (farray_scale*np.array([float(x) for x in work])).round()
    return b64encode(x.astype(np.int32).tobytes()).decode('ascii')

def decode_farray(dat):
    """decodes a string encoded by encode_farray()
    returns list of string
    """
    arr = np.fromstring(b64decode(dat), dtype=np.int32)/farray_scale
    out = []
    for a in arr:
        if (abs(a-2.0) < 1.e-5):
            out.append('?')
        elif (abs(a-3.0) < 1.e-5):
            out.append('.')
        else:
            out.append(f"{a:f}")
    return out

def put_optarray(dat, attr):
    d = dat.get(attr, '0')
    if d != '0':
        d = encode_farray(d)
    return d

def get_optarray(dat):
    if dat not in (0, '0'):
        dat = decode_farray(dat)
    return dat


schema = (
    '''CREATE TABLE version (id integer primary key, tag text, date text, notes text);''',
    '''CREATE TABLE elements (
        id  integer not null,
	z INTEGER NOT NULL,
	name VARCHAR(40),
	symbol VARCHAR(2) NOT NULL primary key);''',

    '''CREATE TABLE spacegroups (
        id INTEGER primary key,
	hm_notation VARCHAR(16) not null unique,
	symmetry_xyz text NOT NULL,
	category text     );''',

    '''CREATE TABLE minerals (
	id INTEGER not null primary key,
	name text not null unique);''',

    '''CREATE TABLE authors (
        id INTEGER NOT NULL primary key,
	name text unique);''',
    '''CREATE TABLE publications (
	id INTEGER NOT NULL primary key,
	journalname text not null,
	volume text,
	year  integer not null,
	page_first text,
	page_last text);''',

    '''CREATE TABLE publication_authors (
        publication_id INTEGER not null,
	author_id integer not null,
	FOREIGN KEY(publication_id) REFERENCES publications (id),
	FOREIGN KEY(author_id) REFERENCES authors (id));''',

    '''CREATE TABLE cif (
        id integer not null primary key,
	mineral_id INTEGER,
	spacegroup_id INTEGER,
	publication_id INTEGER,
	formula text,
        compound text,
        pub_title text,
        formula_title text,
	a text,
	b text,
	c text,
	alpha text,
	beta text,
	gamma text,
        cell_volume text,
        crystal_density text,
	atoms_sites text,
        atoms_x text,
        atoms_y text,
        atoms_z text,
        atoms_occupancy text,
        atoms_u_iso text,
        atoms_aniso_label text,
        atoms_aniso_u11 text,
        atoms_aniso_u22 text,
        atoms_aniso_u33 text,
        atoms_aniso_u12 text,
        atoms_aniso_u13 text,
        atoms_aniso_u23 text,
	qdat text,
     	amcsd_url text,
	FOREIGN KEY(spacegroup_id) REFERENCES spacegroups (id),
	FOREIGN KEY(mineral_id) REFERENCES minerals (id),
	FOREIGN KEY(publication_id) REFERENCES publications (id));''',

    '''CREATE TABLE cif_elements (
        cif_id text not null,
	element VARCHAR(2) not null);''',
    )


atsyms = ['H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg',
        'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V',
        'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se',
        'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh',
        'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba',
        'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho',
        'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt',
        'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac',
        'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm',
        'Md', 'No', 'Lr', 'D']


atnames = ['hydrogen', 'helium', 'lithium', 'beryllium',
         'boron', 'carbon', 'nitrogen', 'oxygen', 'fluorine', 'neon',
         'sodium', 'magnesium', 'aluminum', 'silicon', 'phosphorus',
         'sulfur', 'chlorine', 'argon', 'potassium', 'calcium', 'scandium',
         'titanium', 'vanadium', 'chromium', 'manganese', 'iron', 'cobalt',
         'nickel', 'copper', 'zinc', 'gallium', 'germanium', 'arsenic',
         'selenium', 'bromine', 'krypton', 'rubidium', 'strontium',
         'yttrium', 'zirconium', 'niobium', 'molybdenum', 'technetium',
         'ruthenium', 'rhodium', 'palladium', 'silver', 'cadmium',
         'indium', 'tin', 'antimony', 'tellurium', 'iodine', 'xenon',
         'cesium', 'barium', 'lanthanum', 'cerium', 'praseodymium',
         'neodymium', 'promethium', 'samarium', 'europium', 'gadolinium',
         'terbium', 'dysprosium', 'holmium', 'erbium', 'thulium',
         'ytterbium', 'lutetium', 'hafnium', 'tantalum', 'tungsten',
         'rhenium', 'osmium', 'iridium', 'platinum', 'gold', 'mercury',
         'thallium', 'lead', 'bismuth', 'polonium', 'astatine', 'radon',
         'francium', 'radium', 'actinium', 'thorium', 'protactinium',
         'uranium', 'neptunium', 'plutonium', 'americium', 'curium',
         'berkelium', 'californium', 'einsteinium', 'fermium',
         'mendelevium', 'nobelium', 'lawrencium', 'deuterium' ]

def create_amscifdb(dbname='test.db'):
    if os.path.exists(dbname):
        os.unlink(dbname)

    conn = sqlite3.connect(dbname)
    cursor = conn.cursor()
    for s in schema:
        cursor.execute(s)

    cursor.execute('insert into version values (?,?,?,?)',
                   ('0', 'in progress', 'today', 'in progress'))

    atz, i = 0, 0
    for sym, name in zip(atsyms, atnames):
        i += 1
        atz += 1
        if sym == 'D':
            atz = 1
        cursor.execute('insert into elements values (?,?,?,?)', (i, atz, sym, name))



#!/usr/bin/env python
__date__    = '2022-Jan-15'
__release_version__ = '0.9.58'
__authors__ = "M. Newville, M. Koker, B. Ravel, and others"

import sys
import numpy
import scipy
import matplotlib
import lmfit
from collections import OrderedDict, namedtuple
from packaging.version import parse as ver_parse
try:
    from importlib.metadata import version, PackageNotFoundError
except:
    from importlib_metadata import version, PackageNotFoundError
try:
    __version__ = version("xraylarch")
except PackageNotFoundError:
    # package is not installed
    __version__ = __release_version__

import requests
from requests.packages.urllib3.exceptions import InsecureRequestWarning

def version_data(mods=None):
    sysvers = sys.version
    if '\n' in sysvers:
        sysvers = sysvers.split('\n')[0]

    vdat = OrderedDict()
    vdat['larch'] = f'{__release_version__} ({__date__}) {__authors__}'
    vdat['python'] = "%s" % (sysvers)

    allmods = [numpy, scipy, matplotlib, lmfit]
    if mods is not None:
        for m in mods:
            if m not in allmods:
                allmods.append(m)

    for mod in allmods:
        if mod is not None:
            mname = mod.__name__
            try:
                vers = mod.__version__
            except:
                vers = "unavailable"
            vdat[mname] = vers
    return vdat

def make_banner(mods=None):
    vdat = version_data(mods=mods)

    lines = ['Larch %s' % vdat.pop('larch'),
             'Python %s' % vdat.pop('python')]

    reqs = []
    for name, vstr in vdat.items():
        reqs.append('%s %s' % (name, vstr))
    lines.append(', '.join(reqs))
    if __version__ != __release_version__:
        lines.append(f'Devel Version: {__version__:s}')

    linelen = max([len(line) for line in lines])
    border = '='*max(linelen, 75)
    lines.insert(0, border)
    lines.append(border)

    return '\n'.join(lines)


########
## for comparing current with remote version
########
VERSION_URL='https://raw.githubusercontent.com/xraypy/xraylarch/gh-pages/version.txt'

VersionStatus = namedtuple('VersionStatus', ('update_available',
                                             'local_version',
                                             'remote_version',
                                             'message'))

UPDATE_MESSAGE = """#=== Update Available ===
Larch version {remote_version:s} is available. Your version is currently {local_version:s}.
To update the latest version run
   larch -u
from a Command Window or Terminal.
#========================"""

LATEST_MESSAGE = """Larch version {local_version:s} is up to date."""

def check_larchversion():
    local_version = __release_version__

    requests.packages.urllib3.disable_warnings(InsecureRequestWarning)
    try:
        req = requests.get(VERSION_URL, verify=False, timeout=3)
    except:
        return VersionStatus(False, local_version, 'unknown', 'offline')
    remote_version = '0.9.001'
    if req.status_code == 200:
        try:
            for line in req.text.split('\n'):
                line = line.strip()
                if not line.startswith('#'):
                    remote_version = line
                    break
        except:
            pass
    update_available = ver_parse(remote_version) > ver_parse(local_version)
    message = UPDATE_MESSAGE if update_available else LATEST_MESSAGE
    message = message.format(remote_version=remote_version,
                             local_version=local_version)
    return VersionStatus(update_available,local_version, remote_version, message)



#!/usr/bin/env python
'''
Diffraction functions require for fitting and analyzing data.

mkak 2017.02.06 (originally written spring 2016)
'''

##########################################################################
# IMPORT PYTHON PACKAGES

import math
import numpy as np
from numpy import cos, sin, arcsin,  degrees
from ..utils.physical_constants import PLANCK_HC, TAU, DEG2RAD, RAD2DEG

import re
import os
import cmath

##########################################################################
# FUNCTIONS

def d_from_q(q):
    '''
    Converts q axis into d (returned units inverse of provided units)
    d = 2*PI/q
    '''
    return TAU/q

def d_from_twth(twth,wavelength,ang_units='degrees'):
    '''
    Converts 2th axis into d (returned units same as wavelength units)
    d = lambda/[2*sin(2th/2)]

    ang_unit : default in degrees; will convert from 'rad' if given
    '''
    if not ang_units.startswith('rad'):
        twth = DEG2RAD*twth
    return wavelength/(2*sin(twth/2.))


def twth_from_d(d,wavelength,ang_units='degrees'):
    '''
    Converts d axis into 2th (d and wavelength must be in same units)
    2th = 2*sin^-1(lambda/[2*d])

    ang_unit : default in degrees; will convert to 'rad' if given
    '''
    twth = 2*arcsin(wavelength/(2.*d))
    if ang_units.startswith('rad'):
        return twth
    else:
        return RAD2DEG*twth


def twth_from_q(q,wavelength,ang_units='degrees'):
    '''
    Converts q axis into 2th (q and wavelength will have inverse units)
    2th = 2*sin^-1(lambda/[2*d])

    ang_unit : default in degrees; will convert to 'rad' if given
    '''
    twth = 2*arcsin((q*wavelength)/(2*TAU))
    if ang_units.startswith('rad'):
        return twth
    else:
        return RAD2DEG*twth

def q_from_d(d):
    '''
    Converts d axis into q (returned units inverse of provided units)
    q = 2*PI/d
    '''
    return TAU/d


def q_from_twth(twth, wavelength, ang_units='degrees'):
    '''
    Converts 2th axis into q (q returned in inverse units of wavelength)
    q = [(4*PI)/lamda]*sin(2th/2)

    ang_unit : default in degrees; will convert from 'rad' if given
    '''
    if not ang_units.startswith('rad'):
        twth = DEG2RAD*twth
    return ((2*TAU)/wavelength)*sin(twth/2.)

def qv_from_hkl(hklall, a, b, c, alpha, beta, gamma):

    qv = np.zeros(np.shape(hklall))
    uvol = unit_cell_volume(a,b,c,alpha,beta,gamma)
    alpha, beta, gamma = DEG2RAD*alpha, DEG2RAD*beta, DEG2RAD*gamma
    q0 = [(b*c*sin(alpha))/uvol,(c*a*sin(beta))/uvol,(a*b*sin(gamma))/uvol]

    for i, hkl in enumerate(hklall):
        qv[i] = [TAU*hkl[0]*q0[0], TAU*hkl[1]*q0[1], TAU*hkl[2]*q0[2]]
    return qv

def d_from_hkl(hkl, a, b, c, alpha, beta, gamma):
    h, k, l = hkl[:, 0], hkl[:, 1], hkl[:, 2]
    alpha, beta, gamma = DEG2RAD*alpha, DEG2RAD*beta, DEG2RAD*gamma
    x = 1-cos(alpha)**2 - cos(beta)**2 - cos(gamma)**2 \
        + 2*cos(alpha)*cos(beta)*cos(gamma)
    y = (h*sin(alpha)/a)**2 + 2*k*l*(cos(beta)*cos(gamma)-cos(alpha))/(b*c) + \
        (k*sin(beta)/b)**2 + 2*l*h*(cos(gamma)*cos(alpha)-cos(beta))/(c*a) + \
        (l*sin(gamma)/c)**2 + 2*h*k*(cos(alpha)*cos(beta)-cos(gamma))/(a*b)
    d = np.sqrt(x/y)
    return d

def d_from_hkl_orig(hklall, a, b, c, alpha, beta, gamma):
    d = np.zeros(len(hklall))
    alpha, beta, gamma = DEG2RAD*alpha, DEG2RAD*beta, DEG2RAD*gamma
    for i,hkl in enumerate(hklall):
        h,k,l = hkl
        x = 1-cos(alpha)**2 - cos(beta)**2 - cos(gamma)**2 \
            + 2*cos(alpha)*cos(beta)*cos(gamma)
        y = (h*sin(alpha)/a)**2 + 2*k*l*(cos(beta)*cos(gamma)-cos(alpha))/(b*c) \
            + (k*sin(beta)/b)**2 + 2*l*h*(cos(gamma)*cos(alpha)-cos(beta))/(c*a) \
            + (l*sin(gamma)/c)**2 + 2*h*k*(cos(alpha)*cos(beta)-cos(gamma))/(a*b)
        d[i] = np.sqrt(x/y)
    return d

def unit_cell_volume(a, b, c, alpha, beta, gamma):
    alpha, beta, gamma = DEG2RAD*alpha, DEG2RAD*beta, DEG2RAD*gamma
    return a*b*c*(1-cos(alpha)**2-cos(beta)**2-cos(gamma)**2+
                  2*cos(alpha)*cos(beta)*cos(gamma))**0.5


def E_from_lambda(wavelength, E_units='keV', lambda_units='A'):
    '''
    Converts lambda into energy
    E = hf ; E = hc/lambda

    E_units      : default keV; can convert to 'eV' if given
    lambda_units : default 'A'; can convert from 'm' or 'nm' if given
    '''
    if lambda_units == 'm':
        wavelength = wavelength*1e10
    elif lambda_units == 'nm':
        wavelength = wavelength*10
    if E_units.lower() == 'kev':
        return PLANCK_HC/wavelength*1e-3 # keV
    else:
        return (PLANCK_HC/wavelength)    # eV


def lambda_from_E(E, E_units='keV', lambda_units='A'):
    '''
    Converts lambda into energy
    E = hf ; E = hc/lambda

    E_units      : default keV; can convert from 'eV' if given
    lambda_units : default 'A'; can convert to 'm' or 'nm' if given
    '''
    if E_units.lower() != 'kev':
        E = E*1e-3 # keV
    scale = 1.e-3
    if lambda_units == 'm':
        scale = 1.e-13
    elif lambda_units == 'nm':
        scale = 1e-4
    return scale*PLANCK_HC/E

def generate_hkl(hmax=15, kmax=15, lmax=15, positive_only=True):
    if positive_only:
        hklall = np.mgrid[0:hmax+1, 0:kmax+1, 0:lmax+1].reshape(3, -1).T
    else:
        hklall = np.mgrid[-hmax:hmax+1, -kmax:kmax+1, -lmax:lmax+1].reshape(3, -1).T
    return np.array([hkl for hkl in hklall if hkl[0]**2 + hkl[1]**2 + hkl[2]**2 > 0])



import os
import sys
import time
import logging
import shutil
from datetime import datetime, timedelta
import wx
import wx.lib.scrolledpanel as scrolled
import wx.dataview as dv

import larch
from larch.site_config import user_larchdir
from larch.utils import unixpath
from larch.wxlib import (GridPanel, GUIColors, Button, pack, SimpleText,
                         FileOpen, FileSave, Font, LEFT, FRAMESTYLE,
                         FONTSIZE, MenuItem, EditableListBox, OkCancel,
                         FileCheckList, Choice, HLine, ReportFrame, Popup,
                         LarchWxApp)

from larch.xafs import get_feff_pathinfo
from larch.xray import atomic_symbols

ATSYMS = ['< All >'] + atomic_symbols
EDGES  = ['< All >', 'K', 'L3', 'L2', 'L1', 'M5']


LEFT = LEFT|wx.ALL
DVSTYLE = dv.DV_VERT_RULES|dv.DV_ROW_LINES|dv.DV_MULTIPLE

class FeffPathsModel(dv.DataViewIndexListModel):
    def __init__(self, feffpaths, with_use=True):
        dv.DataViewIndexListModel.__init__(self, 0)
        self.data = []
        self.paths = {}
        self.with_use = with_use
        self.feffpaths = feffpaths
        self.read_data()

    def set_data(self, feffpaths):
        self.paths = {}
        self.feffpaths = feffpaths
        self.read_data()

    def read_data(self):
        self.data = []
        if self.feffpaths is None:
            row = ['feffNNNN.dat', '0.0000', '2', '6', '100.0']
            if self.with_use: row.append(False)
            row.append('* -> * -> *')
            self.data.append(row)
        else:
            for fp in self.feffpaths:
                row = [fp.filename, '%.4f' % fp.reff,
                       '%.0f' % fp.nleg, '%.0f' % fp.degeneracy,
                       '%.3f' % fp.cwratio]
                use = False
                if self.with_use:
                    if fp.filename in self.paths:
                        use = self.paths[fp.filename]
                    row.append(use)
                row.append(fp.geom)
                self.data.append(row)
                self.paths[fp.filename] = use
        self.Reset(len(self.data))


    def select_all(self, use=True):
        for pname in self.paths:
            self.paths[pname] = use
        self.read_data()

    def select_above(self, item):
        itemname = self.GetValue(item, 0)
        use = True
        for row in self.data:
            self.paths[row[0]] = use
            if row[0] == itemname:
                use = not use
        self.read_data()

    def GetColumnType(self, col):
        if self.with_use and col == 5:
            return "bool"
        return "string"

    def GetValueByRow(self, row, col):
        return self.data[row][col]

    def SetValueByRow(self, value, row, col):

        self.data[row][col] = value
        return True

    def GetColumnCount(self):
        return len(self.data[0])

    def GetCount(self):
        return len(self.data)

    def GetAttrByRow(self, row, col, attr):
        """set row/col attributes (color, etc)"""
        nleg = self.data[row][2]
        cname = self.data[row][0]
        if nleg == '2':
            attr.SetColour('#000')
            attr.SetBold(False)
            return True
        elif nleg == '3':
            attr.SetColour('#A11')
            attr.SetBold(False)
            return True
        elif nleg == '4':
            attr.SetColour('#11A')
            attr.SetBold(False)
            return True
        else:
            attr.SetColour('#393')
            attr.SetBold(False)
            return True
        return False


class RemoveFeffCalcDialog(wx.Dialog):
    """dialog for removing Feff Calculations"""

    def __init__(self, parent, ncalcs=1, **kws):
        title = "Remove Feff calculations?"
        wx.Dialog.__init__(self, parent, wx.ID_ANY, title=title, size=(325, 275))
        panel = GridPanel(self, ncols=3, nrows=4, pad=2, itemstyle=LEFT)

        panel.Add(SimpleText(panel, f'Remove {ncalcs:d} Feff calculations?'),
                             dcol=3, newrow=True)
        panel.Add(SimpleText(panel, 'Warning: this cannot be undone!'),
                             dcol=3, newrow=True)
        panel.Add((5, 5), newrow=True)
        panel.Add(HLine(panel, size=(500, 3)), dcol=2, newrow=True)
        panel.Add(OkCancel(panel), dcol=2, newrow=True)
        panel.pack()

    def GetResponse(self):
        self.Raise()
        return (self.ShowModal() == wx.ID_OK)

class FeffResultsPanel(wx.Panel):
    """ present Feff results """
    def __init__(self, parent=None, feffresult=None, path_importer=None, xasmain=None,
                 _larch=None):
        wx.Panel.__init__(self, parent, -1, size=(650, 650))
        self.parent = parent
        self.xasmain = xasmain
        self.path_importer = path_importer
        self._larch = _larch
        self.feffresult = feffresult
        self.report_frame = None

        self.dvc = dv.DataViewCtrl(self, style=DVSTYLE)
        self.dvc.SetMinSize((600, 350))

        self.model = FeffPathsModel(None, with_use=callable(path_importer))
        self.dvc.AssociateModel(self.model)

        panel = wx.Panel(self)
        # panel.SetBackgroundColour(GUIColors.bg)

        sizer = wx.GridBagSizer(1, 1)

        bkws = dict(size=(175, -1))
        btn_header = Button(panel, "Show Full Header", action=self.onShowHeader, **bkws)
        btn_feffinp = Button(panel, "Show Feff.inp",   action=self.onShowFeffInp, **bkws)

        if callable(self.path_importer):
            btn_import = Button(panel, "Import Paths",     action=self.onImportPath, **bkws)
            btn_above  = Button(panel, "Select All Above Current", action=self.onSelAbove,  **bkws)
            btn_none   = Button(panel, "Select None",      action=self.onSelNone, **bkws)

        opts = dict(size=(400, -1), style=LEFT)
        self.feff_folder = SimpleText(panel, '',  **opts)
        self.feff_datetime = SimpleText(panel, '',**opts)
        self.feff_header = [SimpleText(panel, '', **opts),
                            SimpleText(panel, '', **opts),
                            SimpleText(panel, '', **opts),
                            SimpleText(panel, '', **opts),
                            SimpleText(panel, '', **opts),
                            SimpleText(panel, '', **opts)]


        ir = 0
        sizer.Add(SimpleText(panel, 'Feff Folder:'), (ir, 0), (1, 1),  LEFT, 2)
        sizer.Add(self.feff_folder,                  (ir, 1), (1, 4),  LEFT, 2)
        ir += 1
        sizer.Add(SimpleText(panel, 'Date Run:'),    (ir, 0), (1, 1),  LEFT, 2)
        sizer.Add(self.feff_datetime,                (ir, 1), (1, 5),  LEFT, 2)

        ir += 1
        sizer.Add(SimpleText(panel, 'Header:'),      (ir, 0), (1, 1),  LEFT, 1)
        sizer.Add(self.feff_header[0],               (ir, 1), (1, 5),  LEFT, 1)
        ir += 1
        sizer.Add(SimpleText(panel, ''),             (ir, 0), (1, 1),  LEFT, 1)
        sizer.Add(self.feff_header[1],               (ir, 1), (1, 5),  LEFT, 1)
        ir += 1
        sizer.Add(SimpleText(panel, ''),             (ir, 0), (1, 1),  LEFT, 1)
        sizer.Add(self.feff_header[2],               (ir, 1), (1, 5),  LEFT, 1)
        ir += 1
        sizer.Add(SimpleText(panel, ''),             (ir, 0), (1, 1),  LEFT, 1)
        sizer.Add(self.feff_header[3],               (ir, 1), (1, 5),  LEFT, 1)
        ir += 1
        sizer.Add(SimpleText(panel, ''),             (ir, 0), (1, 1),  LEFT, 1)
        sizer.Add(self.feff_header[4],               (ir, 1), (1, 5),  LEFT, 1)
        ir += 1
        sizer.Add(SimpleText(panel, ''),             (ir, 0), (1, 1),  LEFT, 1)
        sizer.Add(self.feff_header[5],               (ir, 1), (1, 5),  LEFT, 1)

        ir += 1
        sizer.Add(btn_header,      (ir, 0), (1, 2),  LEFT, 2)
        sizer.Add(btn_feffinp,     (ir, 2), (1, 2),  LEFT, 2)

        if callable(self.path_importer):
            ir += 1
            sizer.Add(btn_above,     (ir, 0), (1, 2),  LEFT, 2)
            sizer.Add(btn_none,      (ir, 2), (1, 2),  LEFT, 2)
            sizer.Add(btn_import,    (ir, 4), (1, 2),  LEFT, 2)

        ir += 1
        sizer.Add(wx.StaticLine(panel, size=(600, 2)),(ir, 0), (1, 6),  LEFT, 2)

        pack(panel, sizer)

        columns = [('Feff File',   90, 'text'),
                   ('R (\u212B)',  65, 'text'),
                   ('# legs',      60, 'text'),
                   ('# paths',     65, 'text'),
                   ('Importance',  90, 'text')]
        if callable(self.path_importer):
            columns.append(('Use',     50, 'bool'))
        columns.append(('Geometry',   200, 'text'))

        for icol, dat in enumerate(columns):
             label, width, dtype = dat
             method = self.dvc.AppendTextColumn
             mode = dv.DATAVIEW_CELL_EDITABLE
             if dtype == 'bool':
                 method = self.dvc.AppendToggleColumn
                 mode = dv.DATAVIEW_CELL_ACTIVATABLE
             method(label, icol, width=width, mode=mode)
             c = self.dvc.Columns[icol]
             align = wx.ALIGN_RIGHT if icol in (1, 2, 3, 4) else wx.ALIGN_LEFT
             c.Alignment = c.Renderer.Alignment = align
             c.SetSortable(False)


        mainsizer = wx.BoxSizer(wx.VERTICAL)
        mainsizer.Add(panel,    0, LEFT|wx.GROW, 1)
        mainsizer.Add(self.dvc, 1, LEFT|wx.GROW, 1)

        pack(self, mainsizer)
        self.dvc.EnsureVisible(self.model.GetItem(0))

        if feffresult is not None:
            self.set_feffresult(feffresult)

    def onSelAll(self, event=None):
        self.model.select_all(True)

    def onSelNone(self, event=None):
        self.model.select_all(False)

    def onSelAbove(self, event=None):
        if self.dvc.HasSelection():
            self.model.select_above(self.dvc.GetSelection())

    def onShowHeader(self, event=None):
        if self.feffresult is not None:
            self.show_report(self.feffresult.header,
                             title=f'Header for {self.feffresult.folder:s}',
                             default_filename=f'{self.feffresult.folder:s}_header.txt')

    def onShowFeffInp(self, event=None):
        if self.feffresult is not None:
            text = None
            fname = unixpath(os.path.join(self.feffresult.folder, 'feff.inp'))
            if os.path.exists(fname):
                with open(fname, 'rb') as fh:
                    text = fh.read().decode('utf-8')
            else:
                fname = unixpath(os.path.join(user_larchdir, 'feff',
                                              self.feffresult.folder, 'feff.inp'))
                if os.path.exists(fname):
                    with open(fname, 'rb') as fh:
                        text = fh.read().decode('utf-8')
            if text is not None:
                self.show_report(text, title=f'Feff.inp for {self.feffresult.folder:s}',
                                 default_filename=f'{self.feffresult.folder:s}_feff.inp',
                                 wildcard='Input Files (*.inp)|*.inp')

    def show_report(self, text, title='Text', default_filename='out.txt', wildcard=None):
        if wildcard is None:
            wildcard='Text Files (*.txt)|*.txt'
        default_filename = os.path.split(default_filename)[1]
        try:
            self.report_frame.set_text(text)
            self.report_frame.SetTitle(title)
            self.report_frame.default_filename = default_filename
            self.report_frame.wildcard = wildcard
        except:
            self.report_frame = ReportFrame(parent=self,
                                            text=text, title=title,
                                            default_filename=default_filename,
                                            wildcard=wildcard)


    def onImportPath(self, event=None):
        folder  = self.feffresult.folder
        _, fname = os.path.split(folder)
        for data in self.model.data:
            if data[5]:
                self.path_importer(unixpath(os.path.join(folder, data[0])),
                                   self.feffresult)

        self.onSelNone()
        # print(' on import xasmain = ', self.xasmain)
        if self.xasmain is not None:
            self.xasmain.nb.SetSelection(self.xasmain.get_nbpage('feffit')[0])

    def set_feffresult(self, feffresult):
        self.feffresult = feffresult
        self.feff_folder.SetLabel(feffresult.folder)
        self.feff_datetime.SetLabel(feffresult.datetime)
        nhead = len(self.feff_header)

        for i, text in enumerate(feffresult.header.split('\n')[:nhead]):
            self.feff_header[i].SetLabel(text)
        self.model.set_data(feffresult.paths)


class FeffResultsFrame(wx.Frame):
    """ present Feff results """
    def __init__(self,  parent=None, feffresult=None, xasmain=None, _larch=None):
        wx.Frame.__init__(self, parent, -1, size=(850, 650), style=FRAMESTYLE)

        title = "Manage Feff calculation results"
        self.larch = _larch
        if _larch is None:
            self.larch = larch.Interpreter()
        self.larch.eval("# started Feff results browser\n")
        self.larch.eval("if not hasattr('_main', '_feffruns'): _feffruns = {}")
        self.parent = parent
        path_importer = None
        if parent is not None:
            try:
                path_importer = parent.get_nbpage('feffit')[1].add_path
            except:
                pass

        path = unixpath(os.path.join(user_larchdir, 'feff'))
        if not os.path.exists(path):
            os.makedirs(path, mode=493)
        self.feff_folder = path

        self.SetTitle(title)
        self.SetSize((850, 650))
        self.SetFont(Font(FONTSIZE))
        self.createMenus()

        display0 = wx.Display(0)
        client_area = display0.ClientArea
        xmin, ymin, xmax, ymax = client_area
        xpos = int((xmax-xmin)*0.15) + xmin
        ypos = int((ymax-ymin)*0.20) + ymin
        self.SetPosition((xpos, ypos))

        splitter  = wx.SplitterWindow(self, style=wx.SP_LIVE_UPDATE)
        splitter.SetMinimumPaneSize(250)

        # left hand panel
        lpanel = wx.Panel(splitter)
        ltop = wx.Panel(lpanel)

        def Btn(msg, x, act):
            b = Button(ltop, msg, size=(x, 30),  action=act)
            b.SetFont(Font(FONTSIZE))
            return b

        sel_none = Btn('Select None',   120, self.onSelNone)
        sel_all  = Btn('Select All',    120, self.onSelAll)
        tsizer = wx.BoxSizer(wx.HORIZONTAL)
        tsizer.Add(sel_all, 1, LEFT|wx.GROW, 1)
        tsizer.Add(sel_none, 1, LEFT|wx.GROW, 1)
        pack(ltop, tsizer)

        self.fefflist = FileCheckList(lpanel, select_action=self.onShowFeff,
                                      size=(300, -1))

        lsizer = wx.BoxSizer(wx.VERTICAL)
        lsizer.Add(ltop, 0, LEFT|wx.GROW, 1)
        lsizer.Add(self.fefflist, 1, LEFT|wx.GROW|wx.ALL, 1)
        pack(lpanel, lsizer)

        # right hand side
        panel = wx.Panel(splitter)
        wids = self.wids = {}
        toprow = wx.Panel(panel)

        wids['central_atom'] = Choice(toprow, choices=ATSYMS, size=(100, -1),
                                      action=self.onCentralAtom)
        wids['edge']         = Choice(toprow, choices=EDGES, size=(100, -1),
                                      action=self.onAbsorbingEdge)

        flabel = SimpleText(toprow, 'Filter Calculations by Element and Edge:', size=(275, -1))
        tsizer = wx.BoxSizer(wx.HORIZONTAL)
        tsizer.Add(flabel,               1, LEFT, 2)
        tsizer.Add(wids['central_atom'], 0, LEFT|wx.GROW, 2)
        tsizer.Add(wids['edge'],         0, LEFT|wx.GROW, 2)
        pack(toprow, tsizer)

        sizer = wx.BoxSizer(wx.VERTICAL)
        self.feff_panel = FeffResultsPanel(panel, path_importer=path_importer,
                                           xasmain=xasmain, _larch=_larch)
        sizer.Add(toprow, 0, LEFT|wx.GROW|wx.ALL, 2)
        sizer.Add(HLine(panel, size=(550, 2)), 0,  LEFT|wx.GROW|wx.ALL, 2)
        sizer.Add(self.feff_panel, 1, LEFT|wx.GROW|wx.ALL, 2)
        pack(panel, sizer)
        splitter.SplitVertically(lpanel, panel, 1)
        self.Show()
        wx.CallAfter(self.onSearch)

    def onShowFeff(self, event=None):
        fr = self.feffruns.get(self.fefflist.GetStringSelection(), None)
        if fr is not None:
            self.feff_panel.set_feffresult(fr)

    def onSearch(self, event=None):
        catom = self.wids['central_atom'].GetStringSelection()
        edge = self.wids['edge'].GetStringSelection()
        all_catoms = 'All' in catom
        all_edges  = 'All' in edge

        self.fefflist.Clear()
        self.feffruns = {}
        self.larch.eval("## gathering results:\n")
        flist = os.listdir(self.feff_folder)
        flist = sorted(flist, key=lambda t: -os.stat(unixpath(os.path.join(self.feff_folder, t))).st_mtime)

        for path in flist:
            fullpath = unixpath(os.path.join(self.feff_folder, path))
            if os.path.isdir(fullpath):
                try:
                    self.larch.eval(f"_feffruns['{path:s}'] = get_feff_pathinfo('{fullpath:s}')")
                    self.feffruns[path] = thisrun = self.larch.symtable._feffruns[path]
                    if ((all_catoms or (thisrun.absorber == catom)) and
                        (all_edges  or (thisrun.edge == edge))):
                        self.fefflist.Append(path)
                except:
                    print(f"could not read Feff calculation from '{path}'")

    def onCentralAtom(self, event=None):
        self.onSearch()

    def onAbsorbingEdge(self, event=None):
        self.onSearch()

    def onSelAll(self, event=None):
        self.fefflist.select_all()

    def onSelNone(self, event=None):
        self.fefflist.select_none()

    def onRemoveFeffFolders(self, event=None):
        dlg = RemoveFeffCalcDialog(self, ncalcs=len(self.fefflist.GetCheckedStrings()))
        dlg.Raise()
        dlg.SetWindowStyle(wx.STAY_ON_TOP)
        remove = dlg.GetResponse()
        dlg.Destroy()
        if remove:
            for checked in self.fefflist.GetCheckedStrings():
                shutil.rmtree(unixpath(os.path.join(self.feff_folder, checked)))
            self.onSearch()

    def onFeffFolder(self, event=None):
        "prompt for Feff Folder"
        dlg = wx.DirDialog(self, 'Select Main Folder for Feff Calculations',
                           style=wx.DD_DEFAULT_STYLE|wx.DD_CHANGE_DIR)

        dlg.SetPath(self.feff_folder)
        if  dlg.ShowModal() == wx.ID_CANCEL:
            return None
        path = os.path.abspath(dlg.GetPath())
        if not os.path.exists(path):
            os.makedirs(path, mode=493)
        self.feff_folder = path

    def onImportFeffCalc(self, event=None):
        "prompt to import Feff calculation folder"
        dlg = wx.DirDialog(self, 'Select Folder wth Feff Calculations',
                           style=wx.DD_DEFAULT_STYLE|wx.DD_CHANGE_DIR)

        dlg.SetPath(self.feff_folder)
        if  dlg.ShowModal() == wx.ID_CANCEL:
            return None
        path = os.path.abspath(dlg.GetPath())
        if os.path.exists(path):
            flist = os.listdir(path)
            if ('paths.dat' in flist and 'files.dat' in flist and
                'feff0001.dat' in flist and 'feff.inp' in flist):
                _, dname = os.path.split(path)
                dest = unixpath(os.path.join(self.feff_folder, dname))
                shutil.copytree(path, dest)
                self.onSearch()
            else:
                Popup(self, f"{path:s} is not a complete Feff calculation",
                      "cannot import Feff calculation")

    def createMenus(self):
        # ppnl = self.plotpanel
        self.menubar = wx.MenuBar()
        fmenu = wx.Menu()

        MenuItem(self, fmenu, "Rescan Main Feff Folder",
                 "Rescan Feff Folder for Feff calculations",
                 self.onSearch)

        MenuItem(self, fmenu, "Import Feff calculation",
                 "Import other Feff calculation",
                 self.onImportFeffCalc)

        fmenu.AppendSeparator()

        MenuItem(self, fmenu, "Set Main Feff Folder",
                 "Select Main Feff Folder for Feff calculations",
                 self.onFeffFolder)


        MenuItem(self, fmenu, "Remove Selected Feff calculations",
                 "Completely remove Feff calculations",  self.onRemoveFeffFolders)

        fmenu.AppendSeparator()
        MenuItem(self, fmenu, "Quit",  "Exit", self.onClose)

        self.menubar.Append(fmenu, "&File")
        self.SetMenuBar(self.menubar)
        self.Bind(wx.EVT_CLOSE,  self.onClose)

    def onClose(self, event=None):
        self.Destroy()


class FeffResultsBrowserApp(LarchWxApp):
    def __init__(self, dat=None,  **kws):
        self.dat = dat
        LarchWxApp.__init__(self, **kws)

    def createApp(self):
        frame = FeffResultsFrame(feffresult=self.dat)
        self.SetTopWindow(frame)
        return True

if __name__ == '__main__':
    dat = None
    FeffResultsBrowserApp(dat).MainLoop()



#!/usr/bin/env python
#
from __future__ import print_function
import sys
import os
from functools import partial
import wx
import numpy
import scipy

import larch

from wxutils import (MenuItem, Font, Button, Choice)

from .gui_utils import LarchWxApp
from .readlinetextctrl import ReadlineTextCtrl
from .larchfilling import Filling
from .columnframe import ColumnDataFileFrame
from .athena_importer import AthenaImporter
from . import inputhook

from larch.io import (read_ascii, read_xdi, read_gsexdi,
                      gsescan_group, fix_varname,
                      is_athena_project, AthenaProject)
from larch.version import make_banner

FILE_WILDCARDS = "Data Files(*.0*,*.dat,*.xdi)|*.0*;*.dat;*.xdi|All files (*.*)|*.*"

ICON_FILE = 'larch.ico'

def makeColorPanel(parent, color):
    p = wx.Panel(parent, -1)
    p.SetBackgroundColour(color)
    return p

class LarchWxShell(object):
    ps1 = 'Larch>'
    ps2 = ' ... >'
    def __init__(self, wxparent=None,   writer=None, _larch=None,
                 prompt=None, historyfile=None, output=None, input=None):
        self._larch = _larch
        self.textstyle = None
        self.parent = wxparent
        self.prompt = prompt
        self.input  = input
        self.output = output

        if _larch is None:
            self._larch  = larch.Interpreter(historyfile=historyfile,
                                             writer=self)
            self._larch.run_init_scripts()
        self.writer  = self._larch.writer
        self.symtable = self._larch.symtable

        self.objtree = wxparent.objtree

        self.set_textstyle(mode='text')
        self._larch("_sys.display.colors['text2'] = {'color': 'blue'}",
                    add_history=False)

        self.symtable.set_symbol('_builtin.force_wxupdate', False)
        self.symtable.set_symbol('_sys.wx.inputhook',   inputhook)
        self.symtable.set_symbol('_sys.wx.ping',   inputhook.ping)
        self.symtable.set_symbol('_sys.wx.force_wxupdate', False)
        self.symtable.set_symbol('_sys.wx.wxapp', output)
        self.symtable.set_symbol('_sys.wx.parent', wx.GetApp().GetTopWindow())

        if self.output is not None:
            style = self.output.GetDefaultStyle()
            bgcol = style.GetBackgroundColour()
            sfont = style.GetFont()
            self.textstyle = wx.TextAttr('black', bgcol, sfont)

        self.SetPrompt(True)
        # self.flush_timer = wx.Timer(wxparent)
        # self.needs_flush = True
        # wxparent.Bind(wx.EVT_TIMER, self.onFlushTimer, self.flush_timer)
        # wx.CallAfter(self.flush_timer.Start, 500)

    def onUpdate(self, event=None):
        symtable = self.symtable
        if symtable.get_symbol('_builtin.force_wxupdate', create=True):
            app = wx.GetApp()
            evtloop = wx.EventLoop()
            while evtloop.Pending():
                evtloop.Dispatch()
            app.ProcessIdle()
        symtable.set_symbol('_builtin.force_wxupdate', False)


    def SetPrompt(self, complete):
        if self.prompt is None:
            return
        sprompt, scolor = self.ps1, '#000075'
        if not complete:
            sprompt, scolor = self.ps2, '#E00075'
        self.prompt.SetLabel(sprompt)
        self.prompt.SetForegroundColour(scolor)
        self.prompt.Refresh()

    def set_textstyle(self, mode='text'):
        if self.output is None:
            return

        display_colors = self._larch.symtable._sys.display.colors
        textattrs = display_colors.get(mode, {'color':'black'})
        color = textattrs['color']
        style = self.output.GetDefaultStyle()
        bgcol = style.GetBackgroundColour()
        sfont = style.GetFont()
        self.textstyle = wx.TextAttr(color, bgcol, sfont)

    def write_sys(self, text):
        sys.stdout.write(text)
        sys.stdout.flush()

    def write(self, text, **kws):
        if text is None:
            return
        if self.textstyle is None:
            self.set_textstyle()

        if self.output is None or self.textstyle is None:
            self.write_sys(text)
        else:
            self.output.SetInsertionPointEnd()
            pos0 = self.output.GetLastPosition()
            self.output.WriteText(text)
            pos1 = self.output.GetLastPosition()
            self.output.SetStyle(pos0, pos1, self.textstyle)
            self.output.SetInsertionPoint(pos1)
        self.output.Refresh()
        wx.CallAfter(self.input.SetFocus)

    def flush(self, *args):
        self.output.Refresh()
        self.needs_flush = False

    def clear_input(self):
        self._larch.input.clear()
        self.SetPrompt(True)

    def onFlushTimer(self, event=None):
        if self.needs_flush:
            self.flush()

    def eval(self, text, add_history=True, **kws):
        if text is None:
            return
        if text.startswith('!'):
            return os.system(text[1:])

        elif text.startswith('help(') and text.endswith(')'):
            topic = text[5:-1]
            parent = self.symtable.get_parentpath(topic)
            self.objtree.ShowNode("%s.%s" % (parent, topic))
            return
        else:
            if add_history:
                self.parent.AddToHistory(text)
                self.write("%s\n" % text)
            ret = self._larch.eval(text, add_history=add_history)
            if self._larch.error:
                self._larch.input.clear()
                self._larch.writer.set_textstyle('error')
                self._larch.show_errors()
                self._larch.writer.set_textstyle('text')
            elif ret is not None:
                self._larch.writer.write("%s\n" % repr(ret))
            try:
                self.objtree.onRefresh()
            except ValueError:
                pass
            self.SetPrompt(self._larch.input.complete)
            return ret

class LarchPanel(wx.Panel):
    """Larch Input/Output Panel + Data Viewer as a wx.Panel,
    suitable for embedding into apps
    """
    def __init__(self,  parent=None, _larch=None,
                 historyfile='history_larchgui.lar', **kwds):
        self.parent = parent
        if not historyfile.startswith(larch.site_config.user_larchdir):
            historyfile = os.path.join(larch.site_config.user_larchdir,
                                       historyfile)

        wx.Panel.__init__(self, parent, -1, size=(750, 725))

        self.splitter = splitter = wx.SplitterWindow(self, style=wx.SP_LIVE_UPDATE)
        splitter.SetMinimumPaneSize(150)
        self.SetBackgroundColour('#E9EEE0')

        self.objtree = Filling(splitter,  rootLabel='_main')

        self.output = wx.TextCtrl(splitter, -1,  '',
                                  style=wx.TE_MULTILINE|wx.TE_RICH|wx.TE_READONLY)

        self.output.CanCopy()
        self.output.SetInsertionPointEnd()
        # self.output.SetDefaultStyle(wx.TextAttr('black', 'white', sfont))

        splitter.SplitHorizontally(self.objtree, self.output, 0.5)

        ipanel = wx.Panel(self)

        self.prompt = wx.StaticText(ipanel, label='Larch>', size=(65,-1),
                                    style=wx.ALIGN_CENTER|wx.ALIGN_RIGHT)

        self.input = wx.TextCtrl(ipanel, value='', size=(525,-1),
                                 style=wx.TE_LEFT|wx.TE_PROCESS_ENTER)

        self.input.Bind(wx.EVT_TEXT_ENTER, self.onText)
        if sys.platform == 'darwin':
            self.input.Bind(wx.EVT_KEY_UP,  self.onChar)
        else:
            self.input.Bind(wx.EVT_CHAR,  self.onChar)

        self.hist_buff = []
        self.hist_mark = 0

        isizer = wx.BoxSizer(wx.HORIZONTAL)
        isizer.Add(self.prompt,  0, wx.BOTTOM|wx.CENTER)
        isizer.Add(self.input,   1, wx.ALIGN_LEFT|wx.EXPAND)

        ipanel.SetSizer(isizer)
        isizer.Fit(ipanel)

        opts = dict(flag=wx.ALL|wx.EXPAND, border=2)
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(splitter,  1, **opts)
        sizer.Add(ipanel, 0, **opts)

        self.SetSizer(sizer)
        self.larchshell = LarchWxShell(wxparent=self,
                                       _larch = _larch,
                                       historyfile=historyfile,
                                       prompt = self.prompt,
                                       output = self.output,
                                       input  = self.input)

        self.objtree.SetRootObject(self.larchshell.symtable)
        root = self.objtree.tree.GetRootItem()
        self.objtree.tree.Expand(root)

    def write_banner(self):
        self.larchshell.set_textstyle('text2')
        self.larchshell.write(make_banner([wx]))
        self.larchshell.write("\n  \n")
        self.larchshell.set_textstyle('text')

    def update(self):
        self.objtree.onRefresh()

    def onText(self, event=None):
        text =  event.GetString()
        self.input.Clear()
        if text.lower() in ('quit', 'exit', 'quit()', 'exit()'):
            self.parent.onExit()
        else:
            wx.CallAfter(self.larchshell.eval, text)

    def onChar(self, event=None):
        key = event.GetKeyCode()

        entry  = self.input.GetValue().strip()
        pos = self.input.GetSelection()
        ctrl = event.ControlDown()

        if key == wx.WXK_RETURN and len(entry) > 0:
            pass
        if key in (wx.WXK_UP, wx.WXK_DOWN):
            if key == wx.WXK_UP:
                self.hist_mark = max(0, self.hist_mark-1)
            else:
                self.hist_mark += 1
            try:
                wx.CallAfter(self.set_input_text, self.hist_buff[self.hist_mark])
            except IndexError:
                wx.CallAfter(self.set_input_text, '')
        event.Skip()

    def set_input_text(self, text):
        self.input.SetValue(text)
        self.input.SetFocus()
        self.input.SetInsertionPointEnd()


    def AddToHistory(self, text=''):
        for tline in text.split('\n'):
            if len(tline.strip()) > 0:
                self.hist_buff.append(tline)
                self.hist_mark = len(self.hist_buff)


class LarchFrame(wx.Frame):
    def __init__(self, parent=None, _larch=None, is_standalone=True,
                 historyfile='history_larchgui.lar', with_inspection=False,
                 exit_on_close=False, with_raise=True, **kwds):

        self.is_standalone = is_standalone
        self.with_inspection = with_inspection
        self.parent = parent
        self.historyfile = historyfile
        self.subframes = {}
        self.last_array_sel = {}

        wx.Frame.__init__(self, parent, -1, size=(750, 725),
                          style= wx.DEFAULT_FRAME_STYLE)
        self.SetTitle('LarchGUI')

        self.font = wx.Font(12, wx.SWISS, wx.NORMAL, wx.BOLD, 0, "")
        self.SetFont(self.font)
        sbar = self.CreateStatusBar(2, wx.CAPTION)

        self.SetStatusWidths([-2,-1])
        self.SetStatusText("Larch initializing...", 0)

        self.mainpanel = LarchPanel(parent=self, _larch=_larch,
                                    historyfile=historyfile)

        self.larchshell = self.mainpanel.larchshell
        self._larch = self.larchshell._larch

        sizer = wx.BoxSizer(wx.VERTICAL)

        sizer.Add(self.mainpanel, 1, wx.ALL|wx.EXPAND)

        self.SetSizer(sizer)

        if parent is None and exit_on_close:
            self.Bind(wx.EVT_CLOSE,  self.onExit)
        else:
             self.Bind(wx.EVT_CLOSE,  self.onClose)
        self.Bind(wx.EVT_SHOW, self.onShow)
        self.BuildMenus()

        # larchdir = larch.site_config.larchdir

        fico = os.path.join(larch.site_config.icondir, ICON_FILE)
        if os.path.exists(fico):
            self.SetIcon(wx.Icon(fico, wx.BITMAP_TYPE_ICO))
        self.mainpanel.write_banner()
        if with_raise:
            self.Raise()

    def Raise(self):
        self.SetStatusText("Ready", 0)
        self.Refresh()
        wx.Frame.Raise(self)

    def BuildMenus(self):
        menuBar = wx.MenuBar()

        fmenu = wx.Menu()
        if self.is_standalone:
            MenuItem(self, fmenu, "&Read Data File\tCtrl+O",
                     "Read Data File", self.onReadData)
        MenuItem(self, fmenu, "&Read and Run Larch Script\tCtrl+R",
                 "Read and Execute a Larch Script", self.onRunScript)
        MenuItem(self, fmenu, "&Save Session History\tCtrl+S",
                 "Save Session History to File", self.onSaveHistory)
        MenuItem(self, fmenu, 'Change Working Directory\tCtrl+W',
                 'Change Directory', self.onChangeDir)
        MenuItem(self, fmenu, 'Clear Input\tCtrl+D',
                 'Clear Input', self.onClearInput)
        MenuItem(self, fmenu, 'Select Font\tCtrl+F',
                 'Select Font', self.onSelectFont)

        if self.with_inspection:
            MenuItem(self, fmenu, 'Show wxPython Inspector\tCtrl+I',
                     'Debug wxPython App', self.onWxInspect)
        fmenu.AppendSeparator()
        MenuItem(self, fmenu, 'Close Display', 'Close display', self.onClose)
        if self.parent is None:
            MenuItem(self, fmenu, 'E&xit', 'End program', self.onExit)

        menuBar.Append(fmenu, '&File')

        _sys = self.larchshell.symtable._sys
        if self.is_standalone and hasattr(_sys, 'gui_apps'):
            appmenu = wx.Menu()
            x_apps = _sys.gui_apps.keys()
            for appname in sorted(x_apps):
                label, creator = _sys.gui_apps[appname]

                MenuItem(self, appmenu, label, label,
                         partial(self.show_subframe,
                                 name=appname, creator=creator))
            menuBar.Append(appmenu, 'Applications')

        hmenu = wx.Menu()
        MenuItem(self, hmenu, '&About',
                 'Information about this program',  self.onAbout)
        menuBar.Append(hmenu, '&Help')
        self.SetMenuBar(menuBar)

    def onSelectFont(self, event=None):
        fdata = wx.FontData()
        fdata.SetInitialFont(self.font)
        dlg = wx.FontDialog(self, fdata)
        if dlg.ShowModal() == wx.ID_OK:
            self.font = dlg.GetFontData().GetChosenFont()
            self.SetFont(self.font)
            self.mainpanel.output.SetFont(self.font)
            self.mainpanel.objtree.SetFont(self.font)
            self.mainpanel.objtree.text.SetFont(self.font)
        dlg.Destroy()

    def onWxInspect(self, event=None):
        wx.GetApp().ShowInspectionTool()

    def onXRFviewer(self, event=None):
        self.larchshell.eval("xrf_plot()")

    def onClearInput(self, event=None):
        self.larchshell.clear_input()

    def onClearInput(self, event=None):
        self.larchshell.clear_input()

    def show_subframe(self, event=None, name=None, creator=None, **opts):
        if name is None or creator is None:
            return
        shown = False
        if name in self.subframes:
            try:
                self.subframes[name].Raise()
                shown = True
            except:
                del self.subframes[name]
        if not shown:
            self.subframes[name] = creator(parent=self,
                                           _larch=self.larchshell._larch,
                                           **opts)
            self.subframes[name].Show()

    def onReadData(self, event=None):
        wildcard = 'Data file (*.dat)|*.dat|All files (*.*)|*.*'
        dlg = wx.FileDialog(self, message='Open Data File',
                            defaultDir=os.getcwd(),
                            wildcard=FILE_WILDCARDS,
                            style=wx.FD_OPEN|wx.FD_CHANGE_DIR)
        path = None
        if dlg.ShowModal() == wx.ID_OK:
            path = os.path.abspath(dlg.GetPath()).replace('\\', '/')
        dlg.Destroy()

        if path is None:
            return

        if is_athena_project(path):
            self.show_subframe(name='athena_import', filename=path,
                               creator=AthenaImporter,
                               read_ok_cb=self.onReadAthenaProject_OK)
        else:
            filedir, filename = os.path.split(path)
            pref = fix_varname((filename + '_'*8)[:8]).replace('.', '_').lower()

            count, maxcount = 1, 9999
            groupname = "%s%3.3i" % (pref, count)
            while hasattr(self.larchshell.symtable, groupname) and count < maxcount:
                count += 1
                groupname = '%s%3.3i' % (pref, count)

            fh = open(path, 'r')
            line1 = fh.readline().lower()
            fh.close()
            reader = read_ascii
            if 'epics stepscan file' in line1:
                reader = read_gsexdi
            elif 'epics scan' in line1:
                reader = gsescan_group
            elif 'xdi' in line1:
                reader = read_xdi

            dgroup = reader(str(path), _larch=self.larchshell._larch)
            dgroup._path = path
            dgroup._filename = filename
            dgroup._groupname = groupname
            self.show_subframe(name='coledit', event=None,
                               creator=ColumnDataFileFrame,
                               filename=path,
                               last_array_sel=self.last_array_sel,
                               read_ok_cb=self.onReadScan_Success)


    def onReadScan_Success(self, script, path, groupname=None, array_sel=None,
                           overwrite=False):
        """ called when column data has been selected and is ready to be used"""
        self.larchshell.eval(script.format(group=groupname, path=path))
        if array_sel is not None:
            self.last_array_sel = array_sel
        self.larchshell.flush()

    def onReadAthenaProject_OK(self, path, namelist):
        """read groups from a list of groups from an athena project file"""
        read_cmd = "_prj = read_athena('{path:s}', do_fft=False, do_bkg=False)"
        self.larchshell.eval(read_cmd.format(path=path))
        dgroup = None
        script = "{group:s} = extract_athenagroup(_prj.{prjgroup:s})"
        for gname in namelist:
            this = getattr(self.larchshell.symtable._prj, gname)
            gid = str(getattr(this, 'athena_id', gname))
            self.larchshell.eval(script.format(group=gid, prjgroup=gname))
        self.larchshell.eval("del _prj")

    def onRunScript(self, event=None):
        wildcard = 'Larch file (*.lar)|*.lar|All files (*.*)|*.*'
        dlg = wx.FileDialog(self, message='Open and Run Larch Script',
                            wildcard=wildcard,
                            style=wx.FD_OPEN|wx.FD_CHANGE_DIR)
        if dlg.ShowModal() == wx.ID_OK:
            fout = os.path.abspath(dlg.GetPath())
            path, fname = os.path.split(fout)
            os.chdir(path)
            text = "run('%s')" % fname
            self.larchshell.write("%s\n" % text)
            wx.CallAfter(self.larchshell.eval, text)
        dlg.Destroy()

    def onSaveHistory(self, event=None):
        wildcard = 'Larch file (*.lar)|*.lar|All files (*.*)|*.*'
        deffile = 'history.lar'
        dlg = wx.FileDialog(self, message='Save Session History File',
                            wildcard=wildcard,
                            defaultFile=deffile,
                            style=wx.FD_SAVE|wx.FD_CHANGE_DIR)
        if dlg.ShowModal() == wx.ID_OK:
            fout = os.path.abspath(dlg.GetPath())
            self._larch.input.history.save(fout, session_only=True)
            self.SetStatusText("Wrote %s" % fout, 0)
        dlg.Destroy()


    def onText(self, event=None):
        text =  event.GetString()
        self.larchshell.write("%s\n" % text)
        self.input.Clear()
        if text.lower() in ('quit', 'exit'):
            self.onExit()
        else:
            self.panel.AddToHistory(text)
            wx.CallAfter(self.larchshell.eval, text)


    def onChangeDir(self, event=None):
        dlg = wx.DirDialog(None, 'Choose a Working Directory',
                           defaultPath = os.getcwd(),
                           style = wx.DD_DEFAULT_STYLE)

        if dlg.ShowModal() == wx.ID_OK:
            os.chdir(dlg.GetPath())
        dlg.Destroy()
        return os.getcwd()

    def onAbout(self, event=None):
        about_msg =  """LarchGui:
        %s""" % (make_banner([wx]))

        dlg = wx.MessageDialog(self, about_msg,
                               "About LarchGui", wx.OK | wx.ICON_INFORMATION)
        dlg.ShowModal()
        dlg.Destroy()

    def onShow(self, event=None):
        if event.Show:
            self.mainpanel.update()

    def onClose(self, event=None):
        try:
            self.Hide()
        except:
            pass

    def onExit(self, event=None, force=False):
        if force:
            ret = wx.ID_YES
        else:
            dlg = wx.MessageDialog(None, 'Really Quit?', 'Question',
                                   wx.YES_NO | wx.NO_DEFAULT | wx.ICON_QUESTION)
            ret = dlg.ShowModal()

        if ret == wx.ID_YES:
            try:
                self._larch.input.history.save()
            except:
                pass
            try:
                self.Destroy()
            except:
                pass
            sys.exit()
        else:
            try:
                event.Veto()
            except:
                pass

class LarchApp(LarchWxApp):
    "simple app to wrap LarchFrame"
    def __init__(self, **kws):
        LarchWxApp.__init__(self, **kws)

    def createApp(self):
        frame = LarchFrame(exit_on_close=True, with_inspection=self.with_inspect)
        frame.Show()
        self.SetTopWindow(frame)
        return True

if __name__ == '__main__':
    LarchApp().MainLoop()



import os
import random
from io import StringIO

HAS_PYMATGEN = False
try:
    from pymatgen.io.cif import CifParser
    from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
    from pymatgen.core import Molecule
    HAS_PYMATGEN = True
except:
    HAS_PYMATGEN = False

from xraydb import atomic_symbol, atomic_number, xray_edge
from larch.io.fileutils import gformat
from larch.utils.strutils import fix_varname, strict_ascii


def get_atom_map(structure):
    """generalization of pymatgen atom map
    Returns:
        dict of ipots
    """
    unique_pot_atoms = []
    all_sites  = []
    for site in structure:
        for elem in site.species.elements:
            if elem.symbol not in unique_pot_atoms:
                unique_pot_atoms.append(elem.symbol)

    atom_map = {}
    for i, atom in enumerate(unique_pot_atoms):
        atom_map[atom] = i + 1
    return atom_map


def read_cif_structure(ciftext):
    """read CIF text, return CIF Structure

    Arguments
    ---------
      ciftext (string):         text of CIF file or name of the CIF file.

    Returns
    -------
      pymatgen Structure object
    """
    if not HAS_PYMATGEN:
        raise ImportError('pymatgen required')
    try:
        cifstructs = CifParser(StringIO(ciftext), site_tolerance=5.e-4)
        parse_ok = True
        file_found = True
    except:
        parse_ok = False
        file_found = False
        if os.path.exists(ciftext):
            file_found = True
            try:
                cifstructs = CifParser(ciftext, site_tolerance=5.e-4)
                parse_ok = True
            except:
                parse_ok = False

    try:
        cstruct = cifstructs.get_structures()[0]
    except:
        raise ValueError('could not get structure from text of CIF file')

    if not parse_ok:
        if not file_found:
            raise FileNotFoundError(f'file {ciftext:s} not found')
        else:
            raise ValueError('invalid text of CIF file')
    return cstruct

def cif_sites(ciftext, absorber=None):
    "return list of sites for the structure"
    cstruct = read_cif_structure(ciftext)
    out = cstruct.sites
    if absorber is not None:
        out = []
        for site in cstruct.sites:
            if site.species_string.lower() == absorber.lower():
                out.append(site)
    return out

def cif2feffinp(ciftext, absorber, edge=None, cluster_size=8.0, absorber_site=1,
                site_index=None, extra_titles=None, version8=True):
    """convert CIF text to Feff8 or Feff6l input file

    Arguments
    ---------
      ciftext (string):         text of CIF file or name of the CIF file.
      absorber (string or int): atomic symbol or atomic number of absorbing element
                                (see Note 1)
      edge (string or None):    edge for calculation (see Note 2)     [None]
      cluster_size (float):     size of cluster, in Angstroms         [8.0]
      absorber_site (int):      index of site for absorber (see Note 3) [1]
      site_index (int or None): index of site for absorber (see Note 4) [None]
      version8 (bool):          whether to write Feff8l input (see Note 5)[True]
    Returns
    -------
      text of Feff input file

    Notes
    -----
      1. absorber is the atomic symbol or number of the absorbing element, and
         must be an element in the CIF structure.
      2. If edge is a string, it must be one of 'K', 'L', 'M', or 'N' edges (note
         Feff6 supports only 'K', 'L3', 'L2', and 'L1' edges). If edge is None,
         it will be assigned to be 'K' for absorbers with Z < 58 (Ce, with an
         edge energy < 40 keV), and 'L3' for absorbers with Z >= 58.
      3. for structures with multiple sites for the absorbing atom, the site
         can be selected by the order in which they are listed in the sites
         list. This depends on the details of the CIF structure, which can be
         found with `cif_sites(ciftext)`, starting counting by 1.
      4. to explicitly state the index of the site in the sites list, use
         site_index (starting at 1!)
      5. if version8 is False, outputs will be written for Feff6l

    """
    cstruct = read_cif_structure(ciftext)

    sgroup = SpacegroupAnalyzer(cstruct).get_symmetry_dataset()
    space_group = sgroup["international"]

    if isinstance(absorber, int):
        absorber   = atomic_symbol(absorber_z)
    absorber_z = atomic_number(absorber)

    if edge is None:
        edge = 'K' if absorber_z < 58 else 'L3'

    edge_energy = xray_edge(absorber, edge).energy
    edge_comment = f'{absorber:s} {edge:s} edge, around {edge_energy:.0f} eV'

    atoms_map = get_atom_map(cstruct)
    if absorber not in atoms_map:
        atlist = ', '.join(atoms_map.keys())
        raise ValueError(f'atomic symbol {absorber:s} not listed in CIF data: ({atlist})')


    site_atoms = {}  # map xtal site with list of atoms occupying that site
    site_tags = {}
    absorber_count = 0
    for sindex, site in enumerate(cstruct.sites):
        site_species = [e.symbol for e in site.species]
        if len(site_species) > 1:
            s_els = [s.symbol for s in site.species.keys()]
            s_wts = [s for s in site.species.values()]
            site_atoms[sindex] = random.choices(s_els, weights=s_wts, k=1000)
            site_tags[sindex] = f'({site.species_string:s})_{1+sindex:d}'
        else:
            site_atoms[sindex] = [site_species[0]] * 1000
            site_tags[sindex] = f'{site.species_string:s}_{1+sindex:d}'
        if absorber in site_species:
            absorber_count += 1
            if absorber_count == absorber_site:
                absorber_index = sindex

    if site_index is not None:
        absorber_index = site_index - 1

    center = cstruct[absorber_index].coords
    sphere = cstruct.get_neighbors(cstruct[absorber_index], cluster_size)
    symbols = [absorber]
    coords = [[0, 0, 0]]
    tags = [f'{absorber:s}_{1+absorber_index:d}']

    for i, site_dist in enumerate(sphere):
        s_index = site_dist[0].index

        site_symbol = site_atoms[s_index].pop()
        tags.append(site_tags[s_index])
        symbols.append(site_symbol)
        coords.append(site_dist[0].coords - center)
    cluster = Molecule(symbols, coords)

    out_text = ['*** feff input generated by xraylarch cif2feff using pymatgen ***']

    if extra_titles is not None:
        for etitle in extra_titles[:]:
            if not etitle.startswith('TITLE '):
                etitle = 'TITLE ' + etitle
            out_text.append(etitle)

    out_text.append(f'TITLE Formula:    {cstruct.composition.reduced_formula:s}')
    out_text.append(f'TITLE SpaceGroup: {space_group:s}')
    out_text.append(f'TITLE # sites:    {cstruct.num_sites}')

    out_text.append('* crystallographics sites: note that these sites may not be unique!')
    out_text.append(f'*     using absorber at site {1+absorber_index:d} in the list below')
    out_text.append(f'*     selected as absorber="{absorber:s}", absorber_site={absorber_site:d}')
    out_text.append('* index   X        Y        Z      species')
    for i, site in enumerate(cstruct):
        fc = site.frac_coords
        species_string = fix_varname(site.species_string.strip())
        marker = '  <- absorber' if  (i == absorber_index) else ''
        out_text.append(f'* {i+1:3d}   {fc[0]:.6f} {fc[1]:.6f} {fc[2]:.6f}  {species_string:s} {marker:s}')

    out_text.extend(['* ', '', ''])

    if version8:
        out_text.append(f'EDGE    {edge:s}')
        out_text.append('S02     1.0')
        out_text.append('CONTROL 1 1 1 1 1 1')
        out_text.append('PRINT   1 0 0 0 0 3')
        out_text.append('EXAFS   20.0')
        out_text.append(f'RPATH   {cluster_size:.2f}')
        out_text.append('*SCF    5.0')

    else:
        edge_index = {'K': 1, 'L1': 2, 'L2': 3, 'L3': 4}[edge]
        out_text.append(f'HOLE    {edge_index:d}  1.0  * {edge_comment:s} (2nd number is S02)')
        out_text.append('CONTROL 1 1 1 0 * phase, paths, feff, chi')
        out_text.append('PRINT   1 0 0 0')
        out_text.append(f'RMAX    {cluster_size:.2f}')

    out_text.extend(['', 'EXCHANGE 0', '',
                     '*  POLARIZATION  0 0 0', '',
                     'POTENTIALS',
                     '*    IPOT  Z   Tag'])

    ipot, z = 0, absorber_z
    out_text.append(f'   {ipot:4d}  {z:4d}   {absorber:s}')
    for tag in atoms_map.keys():
        ipot += 1
        z = atomic_number(tag)
        out_text.append(f'   {ipot:4d}  {z:4d}   {tag:s}')


    at_lines = [(0, cluster[0].x, cluster[0].y, cluster[0].z, 0, absorber, tags[0])]

    for i, site in enumerate(cluster[1:]):
        sym = site.species_string
        ipot = atoms_map[site.species_string]
        dist = cluster.get_distance(0, i+1)
        at_lines.append((dist, site.x, site.y, site.z, ipot, sym, tags[i+1]))

    out_text.append('')
    out_text.append('ATOMS')
    out_text.append(f'*    x         y         z       ipot  tag   distance  site_info')

    acount = 0
    for dist, x, y, z, ipot, sym, tag in sorted(at_lines, key=lambda x: x[0]):
        acount += 1
        if acount > 500:
            break
        sym = (sym + ' ')[:2]
        out_text.append(f'   {x: .5f}  {y: .5f}  {z: .5f} {ipot:4d}   {sym:s}    {dist:.5f}  * {tag:s}')

    out_text.append('')
    out_text.append('* END')
    out_text.append('')
    return strict_ascii('\n'.join(out_text))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Utility wrapper for h5py-like API to Spec files
===================================================

This is a wrapper on top of `silx.io.open` to read Spec_ files via an HDF5-like API.

.. _SPEC: http://www.certif.com/content/spec

Requirements
------------
- silx (http://www.silx.org/doc/silx/latest/modules/io/spech5.html)
"""

__author__ = ["Mauro Rovezzi", "Matt Newville"]
__version__ = "larch_0.9.57"

import os
import copy
import datetime
import six
import collections
import numpy as np
import h5py
from silx.io.utils import open as silx_open
from silx.io.convert import write_to_h5

# from scipy.interpolate import interp1d
# from scipy.ndimage import map_coordinates
# from larch.math.utils import savitzky_golay
from larch import Group
from larch.utils.strutils import bytes2str
from larch.math.normalization import norm1D
from larch.math.deglitch import remove_spikes_medfilt1d

#: Python 3.8+ compatibility
try:
    collectionsAbc = collections.abc
except Exception:
    collectionsAbc = collections

# UTILITIES (the class is below!)


def _str2rng(rngstr, keeporder=True, rebin=None):
    """simple utility to convert a generic string representing a compact
    list of scans to a sorted list of integers

    Parameters
    ----------
    rngstr : string with given syntax (see Example below)
    keeporder : boolean [True], to keep the original order
                keeporder=False turn into a sorted list
    rebin : integer [None], force rebinning of the final range

    Example
    -------
    > _str2rng('100, 7:9, 130:140:5, 14, 16:18:1')
    > [7, 8, 9, 14, 16, 17, 18, 100, 130, 135, 140]

    """
    _rng = []
    for _r in rngstr.split(", "):  # the space is important!
        if len(_r.split(",")) > 1:
            raise NameError("Space after comma(s) is missing in '{0}'".format(_r))
        _rsplit2 = _r.split(":")
        if len(_rsplit2) == 1:
            _rng.append(_r)
        elif len(_rsplit2) == 2 or len(_rsplit2) == 3:
            if len(_rsplit2) == 2:
                _rsplit2.append("1")
            if _rsplit2[0] == _rsplit2[1]:
                raise NameError("Wrong range '{0}' in string '{1}'".format(_r, rngstr))
            if int(_rsplit2[0]) > int(_rsplit2[1]):
                raise NameError("Wrong range '{0}' in string '{1}'".format(_r, rngstr))
            _rng.extend(range(int(_rsplit2[0]), int(_rsplit2[1]) + 1, int(_rsplit2[2])))
        else:
            raise NameError("Too many colon in {0}".format(_r))

    # create the list and return it (removing the duplicates)
    _rngout = [int(x) for x in _rng]

    if rebin is not None:
        try:
            _rngout = _rngout[:: int(rebin)]
        except Exception:
            raise NameError("Wrong rebin={0}".format(int(rebin)))

    def uniquify(seq):
        # Order preserving uniquifier by Dave Kirby
        seen = set()
        return [x for x in seq if x not in seen and not seen.add(x)]

    if keeporder:
        return uniquify(_rngout)
    else:
        return list(set(_rngout))


def _mot2array(motor, acopy):
    """simple utility to generate a copy of an array containing a
    constant value (e.g. motor position)

    """
    a = np.ones_like(acopy)
    return np.multiply(a, motor)


def _make_dlist(dall, rep=1):
    """make a list of strings representing the scans to average

    Parameters
    ----------
    dall : list of all good scans
    rep : int, repetition

    Returns
    -------
    dlist : list of lists of int

    """
    dlist = [[] for d in range(rep)]
    for idx in range(rep):
        dlist[idx] = dall[idx::rep]
    return dlist


def is_specfile(filename):
    """tests whether file may be a Specfile (text or HDF5)"""
    if not os.path.exists(filename):
        return False
    with open(filename, "rb") as fh:
        topbytes = fh.read(10)
    scans = None
    if (
        topbytes.startswith(b"\x89HDF\r")  # HDF5
        or topbytes.startswith(b"#S ")  # partial Spec file (1 scan)
        or topbytes.startswith(b"#F ")  # full Spec file
    ):  # full specscan
        try:
            scans = DataSourceSpecH5(filename)._scans
        except Exception:
            pass
    return scans is not None


def update_nested(d, u):
    """Update a nested dictionary

    From: https://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth
    """
    for k, v in six.iteritems(u):
        dv = d.get(k, {})
        if not isinstance(dv, collectionsAbc.Mapping):
            d[k] = v
        elif isinstance(v, collectionsAbc.Mapping):
            d[k] = update_nested(dv, v)
        else:
            d[k] = v
    return d


# ==================================================================
# CLASS BASED ON SPECH5 (CURRENT/RECOMMENDED)
# ==================================================================
class DataSourceSpecH5(object):
    """Data source utility wrapper for a Spec/BLISS file read as HDF5 object
    via silx.io.open"""

    _file_types = ("Spec", "HDF5")

    def __init__(self, fname=None, logger=None, urls_fmt="silx", verbose=False):
        """init with file name and default attributes

        Parameters
        ----------
        fname : str
            path string of a file that can be read by silx.io.open() [None]
        logger : logging.getLogger() instance
            [None -> larch.utils.logging.getLogger()]
        urls_fmt : str
            how the data are organized in the HDF5 container
            'silx' : default
            'spec2nexus' : as converted by spec2nexus
        verbose : bool [False]
            if True it lowers the logger level to INFO, othewise WARNING by default
        """
        if logger is None:
            from larch.utils.logging import getLogger

            _logger_name = "larch.io.specfile_reader.DataSourceSpecH5"
            self._logger = getLogger(_logger_name, level="WARNING")
        else:
            self._logger = logger

        if verbose:
            self._logger.setLevel("INFO")

        self._fname = fname
        self._sourcefile = None
        self._sourcefile_type = None
        self._scans = None
        self._scans_names = None
        self._scan_n = None
        self._scan_str = None

        self._scan_kws = {  # to get data from scan
            "ax_name": None,
            "to_energy": None,
            "sig_name": None,
            "mon": None,
            "deglitch": None,
            "norm": None,
        }
        self._scangroup = None  # ScanGroup

        self._mots_url = "instrument/positioners"
        self._cnts_url = "measurement"
        self._title_url = "title"
        self._time_start_url = "start_time"
        self._time_end_url = "end_time"
        self._sample_url = "sample/name"
        self._plotcnts_url = "plotselect"
        self._scan_header_url = "instrument/specfile/scan_header"
        self._file_header_url = "instrument/specfile/file_header"
        self._urls_fmt = "silx"

        if urls_fmt == "spec2nexus":
            self._mots_url = "positioners"
            self._cnts_url = "data"
            self._title_url = "title"
            self._urls_fmt = "spec2nexus"
        elif urls_fmt != "silx":
            self._urls_fmt = None
            self._logger.error("'urls_fmt' not understood")
        self.set_group()

        if self._fname is not None:
            self._init_source_file()

    def _init_source_file(self):
        """init source file object"""
        #: source file object (h5py-like)
        try:
            self._sourcefile = silx_open(self._fname)
            for ft in self._file_types:
                if ft in str(self._sourcefile):
                    self._sourcefile_type = ft
            self._scans = self.get_scans()
            self._scans_names = [scn[0] for scn in self._scans]
            try:
                _iscn = 0
                self.set_scan(self._scans[_iscn][0])  # set the first scan at init
                while len(self.get_counters()) == 1:
                    self._logger.warning(f"not enough data in scan {_iscn+1} '{self.get_title()}'")
                    _iscn += 1
                    self.set_scan(self._scans[_iscn][0])
            except Exception as e:
                self._logger.error(e)
        except OSError:
            self._logger.error(f"cannot open {self._fname}")

    def open(self, mode="r"):
        """Open the source file object with h5py in given mode"""
        try:
            self._sourcefile = h5py.File(self._fname, mode)
        except OSError:
            self._logger.error(f"cannot open {self._fname}")
            pass

    def close(self):
        """Close source file silx.io.spech5.SpecH5"""
        self._sourcefile.close()
        self._sourcefile = None

    def get_scangroup(self, scan=None):
        """get current scan group

        Parameters
        ----------
        scan  : str, int, or None
             scan address
        """
        if scan is not None:
            self.set_scan(scan)
        if self._scangroup is None:
            raise AttributeError(
                "Group/Scan not selected -> use 'self.set_scan()' first"
            )
        else:
            return self._scangroup

    def set_group(self, group_url=None):
        """Select group url

        Parameters
        ----------
        group_url : str (optional)
            hdf5 url with respect to / where scans are stored [None -> /scans]

        Returns
        -------
        none: sets attribute self._group_url
        """
        self._group_url = group_url
        if self._group_url is not None:
            self._logger.info(f"selected group {self._group_url}")

    def set_scan(self, scan, scan_idx=1, group_url=None, scan_kws=None):
        """Select a given scan

        Parameters
        ----------
        scan : int or str
            scan number or name
        scan_idx : int (optional)
            scan repetition index [1]
        group_url : str
            hdf5 url with respect to / where scans are stored [None -> /scans]
        scan_kws : None or dict
            additional keyword arguments used to get data from scan

        Returns
        -------
        none: set attributes
            self._scan_n, self._scan_str, self._scan_url, self._scangroup
        """
        if scan_kws is not None:
            self._scan_kws = update_nested(self._scan_kws, scan_kws)
        if scan in self._scans_names:
            self._scan_str = scan
            self._scan_n = self._scans_names.index(scan)
        else:
            scan_n = scan
            if isinstance(scan, str):
                scan_split = scan.split(".")
                scan_n = scan_split[0]
                try:
                    scan_idx = scan_split[1]
                except IndexError:
                    self._logger.warning("'scan_idx' kept at 1")
                    pass
                try:
                    scan_n = int(scan_n)
                    scan_idx = int(scan_idx)
                except ValueError:
                    self._logger.error("scan not selected, wrong 'scan' parameter!")
                    return
            assert isinstance(scan_n, int), "'scan_n' must be an integer"
            assert isinstance(scan_idx, int), "'scan_idx' must be an integer"
            self._scan_n = scan_n
            if self._urls_fmt == "silx":
                self._scan_str = f"{scan_n}.{scan_idx}"
            elif self._urls_fmt == "spec2nexus":
                self._scan_str = f"S{scan_n}"
            else:
                self._logger.error("wrong 'urls_fmt'")
                return
        if group_url is not None:
            self.set_group(group_url)
        if self._group_url is not None:
            self._scan_url = f"{self._group_url}/{self._scan_str}"
        else:
            self._scan_url = f"{self._scan_str}"
        try:
            self._scangroup = self._sourcefile[self._scan_url]
            self._scan_title = self.get_title()
            self._scan_start = self.get_time()
            self._logger.info(
                f"selected scan '{self._scan_url}' | '{self._scan_title}' | '{self._scan_start}'"
            )
        except KeyError:
            self._scangroup = None
            self._scan_title = None
            self._logger.error(f"'{self._scan_url}' is not valid")


    def _list_from_url(self, url_str):
        """Utility method to get a list from a scan url

        .. warning:: the list is **not ordered**

        """
        try:
            return [i for i in self.get_scangroup()[url_str].keys()]
        except Exception:
            self._logger.error(f"'{url_str}' not found -> use 'set_scan' method first")

    # ================== #
    #: READ DATA METHODS
    # ================== #

    def _repr_html_(self):
        """HTML representation for Jupyter notebook"""

        scns = self.get_scans()
        html = ["<table>"]
        html.append("<tr>")
        html.append("<td><b>Scan</b></td>")
        html.append("<td><b>Title</b></td>")
        html.append("<td><b>Start_time</b></td>")
        html.append("</tr>")
        for scn, tlt, sct in scns:
            html.append("<tr>")
            html.append(f"<td>{scn}</td>")
            html.append(f"<td>{tlt}</td>")
            html.append(f"<td>{sct}</td>")
            html.append("</tr>")
        html.append("</table>")
        return "".join(html)

    def get_scans(self):
        """Get list of scans

        Returns
        -------
        list of strings: [['scan.n', 'title', 'start_time'], ... ]
        """
        allscans = []
        for sn in self._sourcefile["/"].keys():
            sg = self._sourcefile[sn]
            try:
                allscans.append(
                    [
                        sn,
                        bytes2str(sg[self._title_url][()]),
                        bytes2str(sg[self._time_start_url][()]),
                    ]
                )
            except KeyError:
                self._logger.error(f"'{sn}' is a datagroup!")
                #go one level below and try take first dataset only
                dt0 = list(sg.keys())[0]
                sgg = sg[dt0]
                try:
                    scname = f"{sn}/{dt0}"
                    allscans.append(
                        [
                            scname,
                            bytes2str(sgg[self._title_url][()]),
                            bytes2str(sgg[self._time_start_url][()]),
                        ]
                    )
                except Exception:
                    self._logger.error(f"{scname} does not have standard title/time URLs")
                    allscans.append([None, None, None])
        return allscans

    def get_motors(self):
        """Get list of all available motors names"""
        return self._list_from_url(self._mots_url)

    def get_scan_motors(self):
        """Get list of motors names actually used in the scan"""
        all_motors = self._list_from_url(self._mots_url)
        counters = self._list_from_url(self._cnts_url)
        return [i for i in counters if i in all_motors]

    def get_counters(self, remove_motors=False):
        """Get list of counters names

        Parameters
        ----------
        remove_motors:  bool [False]
             whether to remove counters that would also be in the motors list
        """
        counters = self._list_from_url(self._cnts_url)
        if remove_motors:
            motors = self._list_from_url(self._mots_url)
            counters = [i for i in counters if i not in motors]
        return counters

    def get_title(self):
        """Get title str for the current scan

        Returns
        -------
        title (str): scan title self._scangroup[self._title_url][()]
        """
        sg = self.get_scangroup()
        return bytes2str(sg[self._title_url][()])

    def get_time(self):
        """Get start time str for the current scan

        Returns
        -------
        start_time (str): scan start time self._scangroup[self._time_start_url][()]
        """
        sg = self.get_scangroup()
        return bytes2str(sg[self._time_start_url][()])

    def get_timestamp(self):
        """Get timestamp from the current scan"""
        dt = np.datetime64(self.get_time())
        return dt.astype(datetime.datetime).timestamp()

    def get_scan_info_from_title(self):
        """Parser to get scan information from title

        Known types of scans
        --------------------
        Generic: <scan_type> <scan_axis> <start> <end> <npoints> <counting_time>
        'Escan' (ESRF-BM30/BM16)
        'Emiscan' (ESRF-BM30/BM16)
        'fscan' (ESRF-ID26)

        Returns
        -------
        iscn : dict of str
            {
             scan_type : "type of scan",
             scan_axis : "scanned axis",
             scan_start : "",
             scan_end : "",
             scan_pts : "",
             scan_ct : "",
            }
        """
        iscn = dict(
            scan_type=None,
            scan_axis=None,
            scan_start=None,
            scan_end=None,
            scan_pts=None,
            scan_ct=None,
        )

        _title = self.get_title()
        if isinstance(_title, np.ndarray):
            _title = np.char.decode(_title)[0]
        _title_splitted = [s for s in _title.split(" ") if not s == ""]
        _scntype = _title_splitted[0]
        iscn.update(dict(scan_type=_scntype))
        try:
            iscn.update(
                dict(
                    scan_axis=_title_splitted[1],
                    scan_start=_title_splitted[2],
                    scan_end=_title_splitted[3],
                    scan_pts=_title_splitted[4],
                    scan_ct=_title_splitted[5],
                )
            )
        except IndexError:
            try:
                iscn.update(
                    dict(
                        scan_start=_title_splitted[1],
                        scan_end=_title_splitted[2],
                        scan_pts=_title_splitted[3],
                        scan_ct=_title_splitted[4],
                    )
                )
            except IndexError:
                pass

        if _scntype == "Escan":
            iscn.update(dict(scan_axis="Energy"))
        if _scntype == "Emiscan":
            iscn.update(dict(scan_axis="Emi_Energy"))
        if _scntype == "fscan":
            iscn.update(dict(scan_axis="mono_energy"))
        return iscn

    def get_scan_axis(self):
        """Get the name of the scanned axis from scan title"""
        iscn = self.get_scan_info_from_title()
        _axisout = iscn["scan_axis"]
        _mots, _cnts = self.get_motors(), self.get_counters()
        if not (_axisout in _mots):
            self._logger.info(f"'{_axisout}' not in (real) motors")
        if not (_axisout in _cnts):
            self._logger.info(f"'{_axisout}' not in counters")
            _axisout = _cnts[0]
            self._logger.info(f"using the first counter: '{_axisout}'")
        return _axisout

    def get_array(self, cnt=0):
        """Get array of a given counter

        Parameters
        ----------
        cnt : str or int
            counter name or index in the list of counters

        Returns
        -------
        array
        """
        sg = self.get_scangroup()
        cnts = self.get_counters()
        if type(cnt) is int:
            cnt = cnts[cnt]
            self._logger.info("Selected counter %s", cnt)
        if cnt in cnts:
            sel_cnt = f"{self._cnts_url}/{cnt}"
            return copy.deepcopy(sg[sel_cnt][()])
        else:
            errmsg = f"'{cnt}' not found in available counters: {cnts}"
            self._logger.error(errmsg)
            raise ValueError(errmsg)

    def get_motor_position(self, mot):
        """Get motor position

        Parameters
        ----------
        mot : str or int
            motor name or index in the list of motors

        Returns
        -------
        value
        """
        sg = self.get_scangroup()
        mots = self.get_motors()
        if type(mot) is int:
            mot = mots[mot]
            self._logger.info(f"Selected motor '{mot}'")
        if mot in mots:
            sel_mot = f"{self._mots_url}/{mot}"
            return copy.deepcopy(sg[sel_mot][()])
        else:
            self._logger.error(f"'{mot}' not found in available motors: {mots}")
            return None

    def get_scan(self, scan=None, datatype=None):
        """Get Larch group for the current scan

        Parameters
        ----------
        scan  : str, int, or None
             scan address
        datatype : str
            type of data, e.g. 'raw', 'xas'

        Returns
        -------
        larch Group with scan data
        """
        scan_group = self.get_scangroup(scan)
        scan_index = self._scan_n
        scan_name = self._scan_str
        all_labels = self.get_counters()
        motor_names = self.get_scan_motors()
        title = self.get_title()
        timestring = self.get_time()
        timestamp = self.get_timestamp()
        path, filename = os.path.split(self._fname)
        axis = self.get_scan_axis()
        array_labels = [axis]
        array_labels.extend([i for i in motor_names if i not in array_labels])
        array_labels.extend([i for i in all_labels if i not in array_labels])

        scan_header = list(scan_group.get(self._scan_header_url, []))
        file_header = list(scan_group.get(self._file_header_url, []))
        file_type = self._sourcefile_type
        header = []
        for scanh in scan_header:
            if scanh.startswith("#CXDI "):
                header.append(scanh[6:].strip())
        out = Group(
            __name__=f"{file_type} file: {filename}, scan: {scan_name}",
            path=path,
            filename=filename,
            datatype=datatype,
            array_labels=array_labels,
            motor_names=motor_names,
            axis=axis,
            scan_index=scan_index,
            scan_name=scan_name,
            title=title,
            header=header,
            scan_header=scan_header,
            file_header=file_header,
            timestring=timestring,
            timestamp=timestamp,
        )

        data = []
        axis_shape = self.get_array(axis).shape
        for label in array_labels:
            arr = self.get_array(label).astype(np.float64)
            if arr.shape == axis_shape:
                setattr(out, label, arr)
                data.append(arr)
            else:
                self._logger.warning(
                    f"'{label}' skipped (shape is different from '{axis}')"
                )
                ipop = array_labels.index(label)
                array_labels.pop(ipop)
        out.data = np.array(data)
        return out

    def get_axis_data(self, ax_name=None, to_energy=None):
        """Get data for the scan axis

        Description
        -----------
        This method returns the data (=label and array) for a given axis of the
        selected scan. It is primarily targeted to a "scanning" axis, but any
        counter can be used. It is possible to control common conversions, like
        Bragg angle to energy.

        Parameters
        ----------
        ax_name : str or None

        to_energy : dict
            Controls the conversion of the signal to energy [None]

            .. note:: Bragg angle assumed in mrad, output in eV

            {
                "bragg_ax": "str",  #: name of counter used for Bragg angale
                "bragg_ax_type": "str",  #: 'motor' or 'counter'
                "bragg_enc_units": float, #: units to convert encoder to deg (bragg_ax should contain 'enc')
            }

        Returns
        -------
        label, data
        """
        if (ax_name is not None) and (ax_name not in self.get_counters()):
            self._logger.error("%s not a counter", ax_name)
            return None, None
        ax_label = ax_name or self.get_scan_axis()
        ax_data = self.get_array(ax_label)
        if to_energy is not None:
            try:
                from sloth.utils.bragg import ang2kev
            except ImportError:
                def ang2kev(theta, d):
                    from larch.utils.physical_constants import PLANCK_HC
                    theta = np.deg2rad(theta)
                    wlen = 2 * d * np.sin(theta)
                    return (PLANCK_HC / wlen) / 1000.

            bragg_ax = to_energy["bragg_ax"]
            bragg_ax_type = to_energy["bragg_ax_type"]
            bragg_d = to_energy["bragg_d"]
            if bragg_ax_type == "counter":
                bragg_deg = self.get_array(bragg_ax).mean()
            elif bragg_ax_type == "motor":
                bragg_deg = self.get_value(bragg_ax)
            else:
                self._logger.error("wrong 'bragg_ax_type' (motor or counter?)")
            if "enc" in bragg_ax:
                bragg_deg = (np.abs(bragg_deg) / to_energy["bragg_enc_units"]) * 360
            ax_abs_deg = bragg_deg + np.rad2deg(ax_data) / 1000.0
            ax_abs_ev = ang2kev(ax_abs_deg, bragg_d) * 1000.0
            ax_data = ax_abs_ev
            ax_label += "_abs_ev"
            self._logger.debug("Converted axis %s", ax_label)
            xmin = ax_data.min()
            xmax = ax_data.max()
            self._logger.info("%s range: [%.3f, %.3f]", ax_label, xmin, xmax)
        return ax_label, ax_data

    def get_signal_data(self, sig_name, mon=None, deglitch=None, norm=None):
        """Get data for the signal counter

        Description
        -----------
        This method returns the data (=label and array) for a given signal of the
        selected scan. It is possible to control normalization and/or deglitching.

        Order followed in the basic processing:
            - raw data
            - divide by monitor signal (+ multiply back by average)
            - deglitch
            - norm

        Parameters
        ----------
        sig_name : str
        mon : dict
            Controls the normalization of the signal by a monitor signal [None]
            {
                "monitor": "str",  #: name of counter used for normalization
                "cps": bool,  #: multiply back to np.average(monitor)
            }
        deglitch : dict
            Controls :func:`larch.math.deglitch.remove_spikes_medfilt1d` [None]
        norm : dict
            Controls the normalization by given method

        Returns
        -------
        label, data
        """
        #: get raw data
        sig_data = self.get_array(sig_name)
        sig_label = sig_name
        #: (opt) divide by monitor signal + multiply back by average
        if mon is not None:
            if isinstance(mon, str):
                mon = dict(monitor=mon, cps=False)
            mon_name = mon["monitor"]
            mon_data = self.get_array(mon_name)
            sig_data /= mon_data
            sig_label += f"_mon({mon_name})"
            if mon["cps"]:
                sig_data *= np.average(mon_data)  #: put back in counts
                sig_label += "_cps"
        #: (opt) deglitch
        if deglitch is not None:
            sig_data = remove_spikes_medfilt1d(sig_data, **deglitch)
            sig_label += "_dgl"
        #: (opt) normalization
        if norm is not None:

            norm_meth = norm["method"]
            sig_data = norm1D(sig_data, norm=norm_meth, logger=self._logger)
            if norm_meth is not None:
                sig_label += f"_norm({norm_meth})"
        self._logger.info("Loaded signal: %s", sig_label)
        return sig_label, sig_data

    def get_curve(
        self,
        sig_name,
        ax_name=None,
        to_energy=None,
        mon=None,
        deglitch=None,
        norm=None,
        **kws,
    ):
        """Get XY data (=curve) for current scan

        Parameters
        ----------
        *args, **kws -> self.get_axis_data() and self.get_signal_data()

        Returns
        -------
        [ax_data, sig_data, label, attrs] : list of [array, array, str, dict]

        """
        ax_label, ax_data = self.get_axis_data(ax_name=ax_name, to_energy=to_energy)
        sig_label, sig_data = self.get_signal_data(
            sig_name, mon=mon, deglitch=deglitch, norm=norm
        )
        label = f"S{self._scan_n}_X({ax_label})_Y{sig_label}"
        attrs = dict(
            xlabel=ax_label,
            ylabel=sig_label,
            label=label,
            ax_label=ax_label,
            sig_label=sig_label,
        )
        return [ax_data, sig_data, label, attrs]

    # =================== #
    #: WRITE DATA METHODS
    # =================== #

    def write_scans_to_h5(
        self,
        scans,
        fname_out,
        scans_groups=None,
        h5path=None,
        overwrite=False,
        conf_dict=None,
    ):
        """Export a selected list of scans to HDF5 file

        .. note:: This is a simple wrapper to
            :func:`silx.io.convert.write_to_h5`

        Parameters
        ----------
        scans : str, list of ints or list of lists (str/ints)
            scan numbers to export (parsed by _str2rng)
            if a list of lists, scans_groups is required
        fname_out : str
            output file name
        scans_groups : list of strings
            groups of scans
        h5path : str (optional)
            path inside HDF5 [None -> '/']
        overwrite : boolean (optional)
            force overwrite if the file exists [False]
        conf_dict : None or dict (optional)
            configuration dictionary saved as '{hdfpath}/.config'
        """
        self._fname_out = fname_out
        self._logger.info(f"output file: {self._fname_out}")
        if os.path.isfile(self._fname_out) and os.access(self._fname_out, os.R_OK):
            self._logger.info(f"output file exists (overwrite is {overwrite})")
            _fileExists = True
        else:
            _fileExists = False

        #: out hdf5 file
        if overwrite and _fileExists:
            os.remove(self._fname_out)
        h5out = h5py.File(self._fname_out, mode="a", track_order=True)

        #: h5path
        if h5path is None:
            h5path = "/"
        else:
            h5path += "/"

        #: write group configuration dictionary, if given
        if conf_dict is not None:
            from silx.io.dictdump import dicttoh5

            _h5path = f"{h5path}.config/"
            dicttoh5(
                conf_dict,
                h5out,
                h5path=_h5path,
                create_dataset_args=dict(track_order=True),
            )
            self._logger.info(f"written dictionary: {_h5path}")

        #: write scans
        def _loop_scans(scns, group=None):
            for scn in scns:
                self.set_scan(scn)
                _scangroup = self._scangroup
                if _scangroup is None:
                    continue
                if group is not None:
                    _h5path = f"{h5path}{group}/{self._scan_str}/"
                else:
                    _h5path = f"{h5path}{self._scan_str}/"
                write_to_h5(
                    _scangroup,
                    h5out,
                    h5path=_h5path,
                    create_dataset_args=dict(track_order=True),
                )
                self._logger.info(f"written scan: {_h5path}")

        if type(scans) is list:
            assert type(scans_groups) is list, "'scans_groups' should be a list"
            assert len(scans) == len(
                scans_groups
            ), "'scans_groups' not matching 'scans'"
            for scns, group in zip(scans, scans_groups):
                _loop_scans(_str2rng(scns), group=group)
        else:
            _loop_scans(_str2rng(scans))

        #: close output file
        h5out.close()


def str2rng_larch(rngstr, keeporder=True):
    """larch equivalent of _str2rng()"""
    return _str2rng(rngstr, keeporder=keeporder)


str2rng_larch.__doc__ = _str2rng.__doc__


def open_specfile(filename):
    return DataSourceSpecH5(filename)


def read_specfile(filename, scan=None):
    """simple mapping of a Spec/BLISS file to a Larch group"""
    df = DataSourceSpecH5(filename)
    return df.get_scan(scan)



import os
from datetime import datetime
from collections import namedtuple

FPathInfo = namedtuple('FeffPathInfo',
                       ('filename', 'reff', 'nleg',
                        'degeneracy', 'cwratio', 'geom'))
class FeffCalcResults:
    def __init__(self, folder, header=None, ipots=None,
                 paths=None, datetime=None, absorber=None,
                 edge=None):
        self.folder = folder
        self.header = header
        self.ipots = ipots
        self.paths = paths
        self.datetime = datetime
        self.absorber = absorber
        self.edge = edge
                          
def get_feff_pathinfo(folder):
    """get list of Feff path info for a Feff folder
    """
    
    fdat = os.path.join(folder, 'files.dat')
    pdat = os.path.join(folder, 'paths.dat')
    if not os.path.exists(fdat) or not os.path.exists(pdat):
        raise ValueError(f'{folder:s} is not a complete Feff folder - run feff?')

    dtime = datetime.fromtimestamp(os.stat(fdat).st_mtime).isoformat()
    with open(fdat, 'r') as fh:
        fdatlines = fh.readlines()

    with open(pdat, 'r') as fh:
        pathslines = fh.readlines()
    
    paths = {}
    header = []
    waiting_for_dashes = True
    for xline in fdatlines:
        xline = xline.strip()
        if xline.startswith('----'):
            waiting_for_dashes = False
        if waiting_for_dashes:
            header.append(xline)
            continue
        if xline.startswith('feff0'):
            w = xline.split()
            index = int(w[0].replace('feff', '').replace('.dat', ''))
            paths[index] = [w[0], w[5], w[4], w[3], w[2]]

    ipots = ['']*12
    waiting_for_dashes = True
    for i, xline in enumerate(pathslines):
        xline = xline.strip()
        if xline.startswith('----'):
            waiting_for_dashes = False
        if waiting_for_dashes:
            continue
        if 'index, nleg' in xline:
            words = xline.split()
            index = int(words[0])
            nleg  = int(words[1])
            elems = []
            for j in range(nleg+1):
                xline = pathslines[j+i+1].strip().replace("'", '')
                if xline.startswith('x   '):
                    continue
                words = xline.split()[:-3]
                ipot = int(words[3])

                if ipot not in ipots:
                    ipots[ipot] = ' '.join(words[4:])
                elems.append(ipot)
            if index in paths:
                paths[index].append(elems)

    ipots = [i for i in ipots if len(i) > 0]
    absorber = ipots[0]
    ipots[0] = '[%s]' % ipots[0]
    opaths = []
    for pindex, pinfo in paths.items():
        pots = [0] + pinfo[5]
        geom =  ' > '.join([ipots[i] for i in pots])
        opaths.append(FPathInfo(filename=pinfo[0],
                                reff=float(pinfo[1]),
                                nleg=int(float(pinfo[2])),
                                degeneracy=float(pinfo[3]),
                                cwratio=float(pinfo[4]),
                                geom=geom))

    # read absorbing shell
    for line in header:
        line = line.strip()
        if (line.startswith('Abs ') and 'Rmt=' in line
            and 'Rnm=' in line and 'shell' in line):
            words= line.replace('shell', '').strip().split()
            edge = words[-1]
        
    return FeffCalcResults(os.path.abspath(folder),
                           absorber=absorber,
                           edge=edge,
                           ipots=ipots,
                           header='\n'.join(header),
                           paths=opaths,
                           datetime=dtime)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
RIXS data reader for beamline BM16 @ ESRF
=========================================

.. note: RIXS stands for Resonant Inelastic X-ray Scattering

.. note: BM16 is FAME-UHD, French CRG beamline

"""
import os
import time
import numpy as np
from larch.io.specfile_reader import DataSourceSpecH5, _mot2array
from silx.io.dictdump import dicttoh5
from larch.utils.logging import getLogger

_logger = getLogger("io_rixs_bm16")


def _parse_header(fname):
    """Get parsed header for the RIXS_###.log file
    Return
    ------
    header : dict
    """
    with open(fname) as f:
        lines = f.read().splitlines()
    header_lines = [line[1:] for line in lines if line[0] == "#"]
    header = {}
    for line in header_lines:
        ls = line.split(": ")
        try:
            k, v = ls[0], ls[1]
        except IndexError:
            pass
        for s in ("START", "END", "STEP"):
            if s in k:
                v = float(v)
        header[k] = v
    return header


def get_rixs_bm16(
    rixs_logfn,
    sample_name=None,
    out_dir=None,
    counter_signal="absF1",
    counter_norm=None,
    interp_ene_in=True,
    save_rixs=False,
):
    """Build RIXS map as X,Y,Z 1D arrays
    Parameters
    ----------
    rixs_logfn : str
        path to the RIXS_###.log file
    sample_name : str, optional ['UNKNOWN_SAMPLE']
    out_dir : str, optional
        path to save the data [None -> data_dir]
    counter_signal : str
        name of the data column to use as signal
    counter_norm : str
        name of the data column to use as normaliztion
    interp_ene_in: bool
        perform interpolation ene_in to the energy step of ene_out [True]
    save_rixs : bool
        if True -> save outdict to disk (in 'out_dir')
    Returns
    -------
    outdict : dict
        {
        '_x': array, energy in
        '_y': array, energy out
        '_z': array, signal
        'writer_name': str,
        'writer_version': str,
        'writer_timestamp': str,
        'filename_all' : list,
        'filename_root': str,
        'name_sample': str,
        'name_scan': str,
        'counter_all': str,
        'counter_signal': str,
        'counter_norm': str,
        'ene_grid': float,
        'ene_unit': str,
        }
    """
    _writer = "get_rixs_bm16"
    _writer_version = "1.5.1"  #: used for reading back in RixsData.load_from_h5()
    _writer_timestamp = "{0:04d}-{1:02d}-{2:02d}_{3:02d}{4:02d}".format(
        *time.localtime()
    )
    header = _parse_header(rixs_logfn)
    if sample_name is None:
        try:
            sample_name = header["SAMPLE_NAME"]
        except Exception:
            sample_name = "UNKNOWN_SAMPLE"
    sfn = header["DATAFILE"]
    scntype = header["RIXS_SCAN_TYPE"]
    data_dir = os.path.join(os.sep, *sfn.split("/")[1:-1])
    if out_dir is None:
        out_dir = data_dir
    ds = DataSourceSpecH5(sfn)

    logobj = np.genfromtxt(rixs_logfn, delimiter=",", comments="#")
    scans = logobj[:, 0]  # list of scan numers
    enes = logobj[:, 1]

    _counter = 0
    for scan, estep in zip(scans, enes):
        scan = int(scan)
        try:
            ds.set_scan(scan)
            escan, sig, lab, attrs = ds.get_curve(counter_signal)
        except Exception:
            _logger.error(f"cannot load scan {scan}!")
            continue
        if scntype == "rixs_et":
            x = _mot2array(estep, escan)
            y = escan
        else:
            x = escan
            y = _mot2array(estep, escan)
        if _counter == 0:
            xcol = x
            ycol = y
            zcol = sig
        else:
            xcol = np.append(xcol, x)
            ycol = np.append(ycol, y)
            zcol = np.append(zcol, sig)
        _counter += 1
        _logger.info(f"Loaded scan {scan}: {estep} eV")

    outdict = {
        "_x": xcol * 1000, #to eV
        "_y": ycol * 1000, #to eV
        "_z": zcol,
        "writer_name": _writer,
        "writer_version": _writer_version,
        "writer_timestamp": _writer_timestamp,
        "counter_signal": counter_signal,
        "counter_norm": counter_norm,
        "sample_name": sample_name,
        "ene_unit": "eV",
        "rixs_header": header,
        "data_dir": data_dir,
        "out_dir": out_dir,
    }

    if save_rixs:
        fnstr = sfn.split("/")[-1].split(".")[0]
        fnout = "{0}_rixs.h5".format(fnstr)
        dicttoh5(outdict, os.path.join(out_dir, fnout))
        _logger.info("RIXS saved to {0}".format(fnout))

    return outdict


if __name__ == "__main__":
    pass



import os
import copy
import numpy as np

import larch
from larch.larchlib import read_config, save_config
from larch.utils import group2dict, unique_name, fix_varname
from larch.wxlib.plotter import last_cursor_pos
from larch.io import fix_varname
PLOTWIN_SIZE = (550, 550)

class XASController():
    """
    class holding the Larch session and doing the processing work for XAS GUI
    """
    config_file = 'xas_viewer.conf'
    def __init__(self, wxparent=None, _larch=None):
        self.wxparent = wxparent
        self.larch = _larch
        if self.larch is None:
            self.larch = larch.Interpreter()
        self.filelist = None
        self.file_groups = {}
        self.fit_opts = {}
        self.group = None
        self.plot_erange = None
        self.groupname = None
        self.report_frame = None
        self.symtable = self.larch.symtable

    def init_larch(self):
        _larch = self.larch
        old_config = read_config(self.config_file)

        config = self.make_default_config()
        for sname in config:
            if old_config is not None and sname in old_config:
                val = old_config[sname]
                if isinstance(val, dict):
                    config[sname].update(val)
                else:
                    config[sname] = val

        for key, value in config.items():
            setattr(self.larch.symtable._sys.xas_viewer, key, value)
        try:
            os.chdir(config['workdir'])
        except:
            pass

    def make_default_config(self):
        """ default config, probably called on first run of program"""
        config = {'chdir_on_fileopen': True,
                  'workdir': os.getcwd()}
        return config

    def get_config(self, key, default=None):
        "get configuration setting"
        confgroup = self.larch.symtable._sys.xas_viewer
        return getattr(confgroup, key, default)

    def save_config(self):
        """save configuration"""
        conf = group2dict(self.larch.symtable._sys.xas_viewer)
        conf.pop('__name__')
        save_config(self.config_file, conf)

    def set_workdir(self):
        self.larch.symtable._sys.xas_viewer.workdir = os.getcwd()

    def write_message(self, msg, panel=0):
        """write a message to the Status Bar"""
        self.wxparent.statusbar.SetStatusText(msg, panel)

    def close_all_displays(self):
        "close all displays, as at exit"
        self.symtable._plotter.close_all_displays()

    def get_display(self, win=1, stacked=False):
        wintitle='Larch XAS Plot Window %i' % win
        # if stacked:
        #     win = 2
        #    wintitle='Larch XAS Plot Window'
        opts = dict(wintitle=wintitle, stacked=stacked, win=win,
                    size=PLOTWIN_SIZE)
        out = self.symtable._plotter.get_display(**opts)
        if win > 1:
            p1 = getattr(self.symtable._plotter, 'plot1', None)
            if p1 is not None:
                try:
                    siz = p1.GetSize()
                    pos = p1.GetPosition()
                    pos[0] += int(siz[0]/4)
                    pos[1] += int(siz[1]/4)
                    out.SetSize(pos)
                    if not stacked:
                        out.SetSize(siz)
                except Exception:
                    pass
        return out

    def get_group(self, groupname=None):
        if groupname is None:
            groupname = self.groupname
            if groupname is None:
                return None
        dgroup = getattr(self.symtable, groupname, None)
        if dgroup is None and groupname in self.file_groups:
            groupname = self.file_groups[groupname]
            dgroup = getattr(self.symtable, groupname, None)
        return dgroup

    def filename2group(self, filename):
        "convert filename (as displayed) to larch group"
        return self.get_group(self.file_groups[str(filename)])

    def merge_groups(self, grouplist, master=None, yarray='mu', outgroup=None):
        """merge groups"""
        cmd = """%s = merge_groups(%s, master=%s,
        xarray='energy', yarray='%s', kind='cubic', trim=True)
        """
        glist = "[%s]" % (', '.join(grouplist))
        outgroup = fix_varname(outgroup.lower())
        if outgroup is None:
            outgroup = 'merged'

        outgroup = unique_name(outgroup, self.file_groups, max=1000)

        cmd = cmd % (outgroup, glist, master, yarray)
        self.larch.eval(cmd)

        if master is None:
            master = grouplist[0]
        this = self.get_group(outgroup)
        master = self.get_group(master)
        if not hasattr(this, 'xasnorm_config'):
            this.xasnorm_config = {}
        this.xasnorm_config.update(master.xasnorm_config)
        this.datatype = master.datatype
        this.xdat = 1.0*this.energy
        this.ydat = 1.0*getattr(this, yarray)
        this.yerr =  getattr(this, 'd' + yarray, 1.0)
        if yarray != 'mu':
            this.mu = this.ydat
        this.plot_xlabel = 'energy'
        this.plot_ylabel = yarray
        return outgroup

    def set_plot_erange(self, erange):
        self.plot_erange = erange

    def copy_group(self, filename, new_filename=None):
        """copy XAS group (by filename) to new group"""
        groupname = self.file_groups[filename]
        if not hasattr(self.larch.symtable, groupname):
            return

        ogroup = self.get_group(groupname)
        ngroup = larch.Group(datatype=ogroup.datatype, copied_from=groupname)
        for attr in dir(ogroup):
            do_copy = True
            if attr in ('xdat', 'ydat', 'i0', 'data' 'yerr',
                        'energy', 'mu'):
                val = getattr(ogroup, attr)*1.0
            elif attr in ('norm', 'flat', 'deriv', 'deconv',
                          'post_edge', 'pre_edge', 'norm_mback',
                          'norm_vict', 'norm_poly'):
                do_copy = False
            else:
                try:
                    val = copy.deepcopy(getattr(ogroup, attr))
                except ValueError:
                    do_copy = False
            if do_copy:
                setattr(ngroup, attr, val)

        if new_filename is None:
            new_filename = filename + '_1'
        ngroup.filename = unique_name(new_filename, self.file_groups.keys())
        ngroup.groupname = unique_name(groupname, self.file_groups.values())
        setattr(self.larch.symtable, ngroup.groupname, ngroup)
        return ngroup

    def get_cursor(self, win=None):
        """get last cursor from selected window"""
        return last_cursor_pos(win=win, _larch=self.larch)

    def plot_group(self, groupname=None, title=None, plot_yarrays=None,
                   new=True, zoom_out=True, **kws):
        ppanel = self.get_display(stacked=False).panel
        newplot = ppanel.plot
        oplot   = ppanel.oplot
        plotcmd = oplot
        viewlims = ppanel.get_viewlimits()
        if new:
            plotcmd = newplot

        dgroup = self.get_group(groupname)
        if not hasattr(dgroup, 'xdat'):
            print("Cannot plot group ", groupname)

        if ((getattr(dgroup, 'plot_yarrays', None) is None or
             getattr(dgroup, 'energy', None) is None or
             getattr(dgroup, 'mu', None) is None)):
            self.process(dgroup)

        if plot_yarrays is None and hasattr(dgroup, 'plot_yarrays'):
            plot_yarrays = dgroup.plot_yarrays

        popts = kws
        path, fname = os.path.split(dgroup.filename)
        if not 'label' in popts:
            popts['label'] = dgroup.plot_ylabel
        zoom_out = (zoom_out or
                  min(dgroup.xdat) >= viewlims[1] or
                  max(dgroup.xdat) <= viewlims[0] or
                  min(dgroup.ydat) >= viewlims[3] or
                  max(dgroup.ydat) <= viewlims[2])

        if not zoom_out:
            popts['xmin'] = viewlims[0]
            popts['xmax'] = viewlims[1]
            popts['ymin'] = viewlims[2]
            popts['ymax'] = viewlims[3]
           

        popts['xlabel'] = dgroup.plot_xlabel
        popts['ylabel'] = dgroup.plot_ylabel
        if getattr(dgroup, 'plot_y2label', None) is not None:
            popts['y2label'] = dgroup.plot_y2label

        plot_extras = None
        if new:
            if title is None:
                title = fname
            plot_extras = getattr(dgroup, 'plot_extras', None)

        popts['title'] = title
        if hasattr(dgroup, 'custom_plotopts'):
            popts.update(dgroup.custom_plotopts)

        narr = len(plot_yarrays) - 1
        for i, pydat in enumerate(plot_yarrays):
            yaname, yopts, yalabel = pydat
            popts.update(yopts)
            if yalabel is not None:
                popts['label'] = yalabel
            popts['delay_draw'] = (i != narr)

            plotcmd(dgroup.xdat, getattr(dgroup, yaname), **popts)
            plotcmd = oplot

        if plot_extras is not None:
            axes = ppanel.axes
            for etype, x, y, opts in plot_extras:
                if etype == 'marker':
                    popts = {'marker': 'o', 'markersize': 4,
                             'label': '_nolegend_',
                             'markerfacecolor': 'red',
                             'markeredgecolor': '#884444'}
                    popts.update(opts)
                    axes.plot([x], [y], **popts)
                elif etype == 'vline':
                    popts = {'ymin': 0, 'ymax': 1.0,
                             'color': '#888888'}
                    popts.update(opts)
                    axes.axvline(x, **popts)
        ppanel.canvas.draw()

